/*                                                                          */
/*                                                                          */
/*                              S P I K E                                   */
/*                                                                          */
/*                  A Spike-Based Neural Network Simulator                  */
/*                                                                          */
/*                   Copyright (C) 1991-1993 Lloyd Watts                    */
/*                                                                          */
/*             Please send comments to lloyd@pcmp.caltech.edu               */
/*                                                                          */
/*                                                                          */
/* This program is free software; you can redistribute it and/or modify     */
/* it under the terms of the GNU General Public License as published by     */
/* the Free Software Foundation.                                            */
/*                                                                          */
/* This program is distributed in the hope that it will be useful,          */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the            */
/* GNU General Public License for more details.                             */
/*                                                                          */
/* You should have received a copy of the GNU General Public License        */
/* along with this program; see the file COPYING.  If not, write to the     */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA,  */
/* or send email to lloyd@pcmp.caltech.edu.                                 */
/*                                                                          */
/*                                                                          */
/*               REVISION HISTORY:                                          */
/*                                                                          */
/* 931217  039   John Edstrom's improvement to xvgr plotting                */
/* 931213  038   support xvgr plotting                                      */
/* 931213  037   generate .plt file for use by gnuplot                      */
/* 931105  036   open sim_file only for reading, not read-write (Kurt)      */
/* 930405  035   better checking about large designs, larger arrays         */
/* 930326  034   more chatty in debug=10 about multiple tonics              */
/* 930321  033   include plot and postscript commands in mathematica file   */
/* 930319  032   only update neuron if synapse changes (faster, small files)*/
/*               pop SYNAPSE_TRANS_OFF events for binary synapses           */
/* 930314  031   checking for valid labels and numbers                      */
/* 930314  030   multiple tonic inputs allowed to same neuron               */
/*               checks for floating tonic and synapse symbol               */
/*               SYNAPSE and TONIC allowed in .spk file                     */
/*               (not just SYNINH, SYNEXC, TONICINH, TONICEXC)              */
/* 930313  029   allows object labels and node names to be the same         */
/* 930313  028   gnuplot output format supported                            */
/* 930303  027   text input form in .spk file                               */
/*               fixed bug in multiple-line display list command            */
/* 930228  026   optional gaussian jitter on the periodic train input       */
/*               labelling of traces on mathematica plots                   */
/* 930225  025   introduce cliplevel, now binary synapse is special case    */
/* 930222  024   display refract supercedes displayrefract                  */
/* 930222  023   Graded synapse option.                                     */
/* 930212  022   Gnu license added                                          */
/* 930126  021   improved error checking in network, copyright notice added */
/* 930125  020   added processing of BUFFER symbols                         */
/* 930123  019   now checking ntk specs in interpret_ntk_file               */
/*               writing allcurves as a part of write_point                 */
/*               open_output_files created                                  */
/*               complete object listing at debug level 4                   */
/*               begun to process BUFFER symbols                            */
/* 930122  018   improved event listing at debug level 2                    */
/* 930122  017   write_point created                                        */
/* 930118  016   creates "allcurves" list for plotting mathematica files    */
/* 930118  015   now writes mathematica-compatible files                    */
/* 930118  014   now flattens hierarchy, checks for "temp" filename         */
/* 930117  013   copying ntk file over, preparing to flatten hierarchy      */
/* 930112  012   removed unneeded dump_event command just before main loop  */
/*               and print out version number                               */
/* 930110  011   adding bursting input, also fixed bug in sorting spike     */
/*                event file                                                */
/* 930110  010   adding poisson input                                       */
/* 930110  009   back to version 005, using external files for spike events */
/*               to allow virtually infinite number of spike events         */
/* 921221  008   fixed wrong range over ENUM types                          */
/* 921221  007   poisson process added, also fixed bug in sorting neuron    */
/*               list (did not clear the current line by overwriting)       */
/* 921221  006   no external files needed in interpret_sim_file             */
/* 921221  005   display list option                                        */
/* 921217  004   display nosynapse option                                   */
/* 921217  003   automatically generated synapse labels                     */
/*               improved debug features                                    */
/* 921217  002   rework of interpret_ntk_file routine, no external files    */
/* 921216  001   debug level is a command-line option                       */
/*               blank lines allowed in .sim file                           */
/*               don't delete temp files in debug mode                      */
/* 921216  000   Starting with Caltech version.                             */
/*                                                                          */
/* 920723  023   threshold option                                           */
/* 920723  022   start_runtime and displayrefract options                   */
/* 920723  021   added tonic input symbol                                   */
/*                                                                          */
/* 910722  020   additional error checking on input files                   */
/* 910720  019   sort objects according to their labels for display         */
/*               clean up files, put view data all in one file.             */
/* 910719  018   interprets .sim file fully.                                */
/* 910719  017   interprets .ntk file fully.                                */
/* 910718  016   taking input from .ntk file.                               */
/* 910714  015   found bug in updating charge_start_value.                  */
/* 910714  014   added pop_specified_event, moved find_next_event into      */
/*               main loop while statement.                                 */
/* 910712  013   modified event handler to key on event-type, not time.     */
/* 910712  012   added dump_object for diagnostic purposes.                 */
/* 910712  011   adjust max_event_index down in pop_event.                  */
/* 910711  010   error checking of input file.                              */
/* 910711  009   keep track of max_event_index to run faster.               */
/* 910711  008   added event counter, push_event().                         */
/* 910711  007   extra plotting offset option.                              */
/* 910710  006   basic event processing completed.                          */
/* 910710  005   option for plotting offset by object number                */
/* 910710  004   View-compatible plotting added.                            */
/* 910710  003   END_RUNTIME command added.                                 */
/* 910710  002   most events processed.  object_map table introduced.       */
/* 910708  001   Created.  Code written up to main event loop.              */
/*                                                                          */
/*                                                                          */
/****************************************************************************/


/* include files */
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <ctype.h>

/* definitions */
#define TRUE              1
#define FALSE             0
#define INFINITY          1.0000000E+9

#define MAX_NUM_ARGS 4
#define MAX_NUM_INPUTS 100
#define MAX_NUM_NEURONS   100
#define MAX_NUM_TONICS   100
#define MAX_NUM_BUFFERS   100
#define MAX_NUM_SYNAPSES  100
#define MAX_NUM_OBJECTS   100
#define MAX_NUM_EVENTS    100
#define MAX_NUM_EXPAND_CELL_PARAMETERS    200
#define MAX_NUM_SYMBOL_TYPES    300
#define MAX_CHARS_PER_LINE 200
#define MAX_TOKENS_PER_LINE 40
#define MAX_FILE_NAME_LENGTH 80
#define DEFAULT_MAX_MATHLABEL_LENGTH 9
#define DEFAULT_RUNTIME 100
#define NUM_EVENT_TYPES   9
#define NUM_OBJECT_TYPES  5
#define NUM_NEURON_STATES  5
#define NUM_SYNAPSE_STATES  2
#define NUM_DISPLAY_TYPES  5
#define NO_DISPLAY_VALUE -99
#define SUFFIX_STRING_DIGITS 3  /* for object filenames */
#define SUFFIX_STRING_DIGITS1 4  /* one more for the \0 */
#define V_THRESHOLD   1.0   /* volts */
#define CAPACITANCE   1.0   /* pF */
#define HIGH_VOLTAGE   5.0
#define REFRACTORY_VOLTAGE -0.7
#define PLOT_MARGIN 0.2
#define TRANS_ON_HEIGHT 0.6

/* for random number generator */
#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define MASK 123459876

/* external procedures */

/* local procedures */
void dump_events();
void dump_map();
void dump_object();
void interpret_ntk_file();
void interpret_spk_file();
void interpret_sim_file();
void check_for_network_errors();
int find_next_event();
void pop_event();
void pop_specified_event();
void push_event();
void open_output_files();
void write_point();
float shift_plot_point();
float ran0();
float gasdev();
int valid_label();
int valid_number();
int found_node_string();

/* global enumeration types */
enum event_enum {NEURON_SPIKE,
                 INPUT_SPIKE,
                 BUFFER_SPIKE,
                 NEURON_REFRACT_OFF,
                 NEURON_ZERO,
                 SYNAPSE_TRANS_ON,
                 SYNAPSE_TRANS_OFF,
                 END_RUNTIME,
                 LAST_EVENT_ENUM};
/* if this is changed, NUM_EVENT_TYPES must also be changed */

enum object_enum {NEURON,
                 SYNAPSE,
                 INPUT,
                 BUFFER,
                 LAST_OBJECT_ENUM};
/* if this is changed, NUM_OBJECT_TYPES must also be changed */

enum neuron_state_enum {LOW_DISCHARGED,
                 CHARGING_UP,
                 CHARGING_DOWN,
                 CHARGED_STEADY,
                 LOW_REFRACTORY};

enum synapse_state_enum {LOW,
                 HIGH};

enum offset_enum {NONE,
                 UP,
                 DOWN,
                 LAST_OFFSET_ENUM};

enum display_type_enum {DEFAULT,
                 ALL,
                 NOSYNAPSE,
                 LIST,
                 LAST_DISPLAY_TYPE_ENUM};

enum ntk_read_state_enum {LOOKING,
                 GOT_C,
                 GOT_CH,
                 GOT_H};

enum output_format_enum {CVIEW,
                 MATHEMATICA,
                 GNUPLOT,
                 XVGR};

enum spk_tokens_enum {SPK_DEF,
                 SPK_ENDDEF,
                 SPK_BEGIN,
                 SPK_END,
                 SPK_OTHER};

enum spk_read_state_enum {SPK_LOOKING,
                 SPK_GOT_DEF,
                 SPK_GOT_BEGIN,
                 SPK_GOT_TOPLEVEL};



/* global variables */
struct event_struct {
  float time;
  int object_id;
  int type;
  int object_sub_id;
} event[MAX_NUM_EVENTS];

struct object_map_struct {
  int type;
  int sub_id;
  char to_node_label[MAX_CHARS_PER_LINE];
  char from_node_label[MAX_CHARS_PER_LINE];
  char label[MAX_CHARS_PER_LINE];
  int display_order;
  int saved_file_num;
} object_map[MAX_NUM_OBJECTS];

struct neuron_struct {
  int id;
  float i_tonic;
  float t_refract;
  int state;
  float charge_rate;
  float charge_start_time;
  float charge_start_value;
  float expected_spike_time;
  float expected_zero_time;
  float last_spike_time;
} neuron[MAX_NUM_NEURONS];

struct tonic_struct {
  int id;
  char to_node_string[MAX_CHARS_PER_LINE];
  char i_tonic_string[MAX_CHARS_PER_LINE];
} tonic[MAX_NUM_TONICS];

struct synapse_struct {
  int id;
  int object_from;
  int object_to;
  float i_weight;
  float t_neuro;
  int state;
  int cliplevel;
  float expected_off_time;
} synapse[MAX_NUM_SYNAPSES];

struct buffer_struct {
  int id;
  int object_from;
  float t_delay;
} buffer[MAX_NUM_BUFFERS];

float        runtime;
float        start_runtime;
float        threshold;
int          debug;
int          displayrefract, output_format;
int          max_event_index;
int          max_mathlabel_length;
char        *neuron_state_string[NUM_NEURON_STATES];
char        *synapse_state_string[NUM_SYNAPSE_STATES];
char        *event_type_string[NUM_EVENT_TYPES];
char        *object_type_string[NUM_OBJECT_TYPES];
FILE        *ntk_file;
FILE        *spk_file;
FILE        *sim_file;
FILE        *evt_file;
FILE        *plt_file;
FILE        *object_file[MAX_NUM_OBJECTS];
char         ntk_file_name[MAX_FILE_NAME_LENGTH];
char         spk_file_name[MAX_FILE_NAME_LENGTH];
char         sim_file_name[MAX_FILE_NAME_LENGTH];
char         evt_file_name[MAX_FILE_NAME_LENGTH];
char         dat_file_name[MAX_FILE_NAME_LENGTH];
char         vgr_file_name[MAX_FILE_NAME_LENGTH];
char         plt_file_name[MAX_FILE_NAME_LENGTH];
char         math_file_name[MAX_FILE_NAME_LENGTH];
char         ps_file_name[MAX_FILE_NAME_LENGTH];
char         basic_file_name[MAX_FILE_NAME_LENGTH];
int          neuron_num,synapse_num,input_num,object_num,offset;
int          tonic_num, num_errors_found;
int          saved_file_num,buffer_num;
int          iset, ntk_file_found, spk_file_found;
float        gset;

/************************************************************************/
void main(argc, argv)
int argc;
char* argv[MAX_NUM_ARGS];
/************************************************************************/

/* BEGIN */
{
    extern struct neuron_struct neuron[];
    extern struct synapse_struct synapse[];
    extern struct event_struct event[];
    extern struct object_map_struct object_map[];
    extern int          debug;
    extern float        runtime;
    extern float        start_runtime;
    extern int          max_event_index;
    extern int          max_mathlabel_length;
    extern int          displayrefract, output_format;
    extern char        *event_type_string[];
    extern char        *object_type_string[];
    extern FILE        *ntk_file;
    extern FILE        *spk_file;
    extern FILE        *sim_file;
    extern FILE        *evt_file;
    extern FILE        *plt_file;
    extern FILE        *object_file[MAX_NUM_OBJECTS];
    extern char         ntk_file_name[MAX_FILE_NAME_LENGTH];
    extern char         spk_file_name[MAX_FILE_NAME_LENGTH];
    extern char         sim_file_name[MAX_FILE_NAME_LENGTH];
    extern char         evt_file_name[MAX_FILE_NAME_LENGTH];
    extern char         dat_file_name[MAX_FILE_NAME_LENGTH];
    extern char         vgr_file_name[MAX_FILE_NAME_LENGTH];
    extern char         plt_file_name[MAX_FILE_NAME_LENGTH];
    extern char         math_file_name[MAX_FILE_NAME_LENGTH];
    extern char         ps_file_name[MAX_FILE_NAME_LENGTH];
    extern char         basic_file_name[MAX_FILE_NAME_LENGTH];
    extern int          neuron_num,synapse_num,input_num,object_num;
    extern int          offset, tonic_num;
    extern int          saved_file_num,buffer_num;
    extern int          ntk_file_found;
    extern int          num_errors_found;


    FILE               *fopen();
    char                network_spec[MAX_CHARS_PER_LINE];
    char                dummy_line[MAX_CHARS_PER_LINE];
    char                command_string[MAX_CHARS_PER_LINE];
    char                dummy_label[MAX_CHARS_PER_LINE];
    char               *ptokens[MAX_TOKENS_PER_LINE];
    char               *source;
    char               *dest;
    int                 i,numtokens;
    int                 event_index;
    int                 dest_index,dest_id;
    int                 event_counter;
    int                 synapse_value_changed;
    int                 jj,kk;
    float               input_sum,write_value;
    float               displayed_refractory_voltage;
    struct event_struct current_event;


    /* print greeting */
    printf("\n  SPIKE 1.039\n  A Spike-Based Neural Network Simulator\n");
    printf("  Copyright (C) 1991-1993 Lloyd Watts\n");
    printf("  This program is distributed under the GNU General Public License\n");
    printf("  Please send comments to lloyd@pcmp.caltech.edu\n\n");
    fflush(stdout);

    /* initialize basic values */
    neuron_num = 0;
    synapse_num = 0;
    input_num = 0;
    object_num = 0;
    buffer_num = 0;
    tonic_num = 0;
    synapse_state_string[0] = "LOW";
    synapse_state_string[1] = "HIGH";
    neuron_state_string[0] = "LOW_DISCHARGED";
    neuron_state_string[1] = "CHARGING_UP";
    neuron_state_string[2] = "CHARGING_DOWN";
    neuron_state_string[3] = "CHARGED_STEADY";
    neuron_state_string[4] = "LOW_REFRACTORY";
    object_type_string[0] = "NEURON";
    object_type_string[1] = "SYNAPSE";
    object_type_string[2] = "INPUT";
    object_type_string[3] = "BUFFER";
    object_type_string[4] = "LAST_OBJECT_ENUM";
    event_type_string[0] = "NEURON_SPIKE";
    event_type_string[1] = "INPUT_SPIKE";
    event_type_string[2] = "BUFFER_SPIKE";
    event_type_string[3] = "NEURON_REFRACT_OFF";
    event_type_string[4] = "NEURON_ZERO";
    event_type_string[5] = "SYNAPSE_TRANS_ON";
    event_type_string[6] = "SYNAPSE_TRANS_OFF";
    event_type_string[7] = "END_RUNTIME";
    event_type_string[8] = "LAST_EVENT_ENUM";
    debug = 0;
    for (i=0; i<MAX_NUM_EVENTS; i++) event[i].type = (int) LAST_EVENT_ENUM;
    for (i=0; i<MAX_NUM_OBJECTS; i++) object_map[i].type=(int) LAST_OBJECT_ENUM;
    offset = (int) DOWN;
    event_counter = 0;
    displayrefract=0;
    displayed_refractory_voltage=0.0;
    threshold=V_THRESHOLD;
    max_mathlabel_length = DEFAULT_MAX_MATHLABEL_LENGTH;
    iset=0; /* for gasdev random number generator */
    ntk_file_found = FALSE;
    spk_file_found = FALSE;
    output_format = (int)CVIEW;
    num_errors_found = 0;
    runtime=(float)DEFAULT_RUNTIME;
    start_runtime=0;

    /* interpret command line */
    /* usage: spike <filename> [-d <level>] */

    /* get filename */
    if (argc <= 1 || argc ==3 || argc > 4) {
      printf("  Usage:  spike <filename> [-d <level>]\n");
      fflush(stdout);
      exit(1);
    }

    /* generate filenames */
    argv++;
    strcpy(ntk_file_name, *argv);

    /* check for filename that start with "temp" */
    if (ntk_file_name[0]=='t' && ntk_file_name[1]=='e' &&
        ntk_file_name[2]=='m' && ntk_file_name[3]=='p') {
      printf("  Usage:  spike <filename> [-d <level>]\n");
      printf("        <filename> may not begin with \"temp\" \n");
      fflush(stdout);
      exit(1);
    }

    strcat(ntk_file_name,".ntk");
    strcpy(spk_file_name, *argv);
    strcat(spk_file_name,".spk");
    strcpy(sim_file_name, *argv);
    strcat(sim_file_name,".sim");
    strcpy(evt_file_name, *argv);
    strcat(evt_file_name,".evt");
    strcpy(dat_file_name, *argv);
    strcat(dat_file_name,".dat");
    strcpy(plt_file_name, *argv);
    strcat(plt_file_name,".plt");
    strcpy(vgr_file_name, *argv);
    strcat(vgr_file_name,".vgr");
    strcpy(math_file_name, *argv);
    strcat(math_file_name,".math");
    strcpy(ps_file_name, *argv);
    strcat(ps_file_name,".ps");
    strcpy(basic_file_name, *argv);

    if (argc == 4) {
      if (strcmp(*++argv,"-d")==0) {
        strcpy(dummy_line,*++argv);
        debug = atoi(&dummy_line[0]);
      } else {
        printf("Usage:  spike <filename> [-d <level>]\n");
        fflush(stdout);
        exit(1);
      }
    }

    /* interpret .ntk file */
    interpret_ntk_file();

    if (!ntk_file_found) {
      interpret_spk_file();
      if (!spk_file_found) {
        printf("\n  Could not find file %s\n  Could not find file %s\n",
           ntk_file_name, spk_file_name);
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }
    }
  
    /* check for network errors */
    check_for_network_errors();

    /* initialize all neurons to LOW_REFRACTORY state */
    for (i=0; i<neuron_num; i++) neuron[i].state = (int) LOW_REFRACTORY;

    /* initialize all synapses */
    for (i=0; i<synapse_num; i++) {
      synapse[i].state = 0;
      synapse[i].expected_off_time = INFINITY;
    }

    printf("  interpreting %s\n",sim_file_name); fflush(stdout);

    /* interpret .sim file */
    interpret_sim_file();

    if (displayrefract==1) displayed_refractory_voltage=REFRACTORY_VOLTAGE;

    /* push NEURON_REFRACT_OFF events for each neuron at t=start_runtime */
    for (i=0; i<neuron_num; i++) {
      push_event(start_runtime, neuron[i].id, (int)NEURON_REFRACT_OFF,i);
    }

    /* push first INPUT_SPIKE event */
    fseek(evt_file, (long) 0, 0);
    if (fgets(network_spec,MAX_CHARS_PER_LINE,evt_file) != NULL) {
      ptokens[0] = (char *) strtok(network_spec," ,\n");
      numtokens=1;
      while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n")) 
              != NULL) numtokens++;
      push_event(atof(ptokens[1]), atoi(ptokens[0]),(int)INPUT_SPIKE,
        object_map[atoi(ptokens[0])].sub_id);
    }

    /* push END_RUNTIME event */
    push_event(runtime, -1, (int)END_RUNTIME,-1);

    if (debug>=1) {
      printf("  opening output data files.\n"); fflush(stdout);
    }

    /* open data file for each object with appropriate header */
	open_output_files();

    if (debug>=1) dump_map();

    if (debug>=4) {
      printf ("              COMPLETE OBJECT LISTING          \n");
      printf ("  ---------------------------------------------\n\n");
      fflush(stdout);
      for (i=0;i<object_num;i++) {
        dump_object(i);
      }
    }

    printf("  beginning simulation.\n");

    if (debug==2) {
      printf ("\n                EVENT LISTING \n\n");
      printf ("  number     time         event type         id\n");
      printf ("  ---------------------------------------------\n");
      fflush(stdout);
    }

    /* MAIN LOOP */
    /* WHILE (events exist within specified time window) DO */
    while ((event_index = find_next_event()) >= 0) {

      if (debug==2) {
        printf("  %4d     %7.3f    %-18s  %5d  \n",
          event_counter+1,event[event_index].time, 
          event_type_string[event[event_index].type],
          event[event_index].object_id);
      }

      if (debug>2) dump_events(event_index);

      /* increment event_counter */
      event_counter++;

      /* store event */
      current_event.time = event[event_index].time;
      current_event.object_id = event[event_index].object_id;
      current_event.type = event[event_index].type;
      current_event.object_sub_id = event[event_index].object_sub_id;

      /* pop the event from the event list */
      pop_specified_event(event_index);

      /* SWITCH (event.type) */
      switch (current_event.type) {

        /* CASE (NEURON_REFRACT_OFF): */
        case NEURON_REFRACT_OFF:

          if (debug>=5) {
            printf("      NEURON_REFRACT_OFF event.\n");
            fflush(stdout);
          }

          if (debug>=5) dump_object(current_event.object_id);

          /* write point */
          write_point(current_event.time,current_event.object_id,
                      0.0, FALSE);

          /* compute sum of inputs for given neuron */
          /* sum of inputs = tonic + sum (synapse_weight*synapse_state) */
          input_sum = neuron[current_event.object_sub_id].i_tonic;
          for (i=0; i<synapse_num; i++) {
            if (synapse[i].object_to == current_event.object_id) {
              if (synapse[i].state != 0) {
                if (synapse[i].state < synapse[i].cliplevel) {
                  input_sum += synapse[i].i_weight*synapse[i].state;
                } else {
                  input_sum += synapse[i].i_weight*synapse[i].cliplevel;
                }
              }
            }
          }

          /* IF (sum of inputs > 0) */
          if (input_sum > 0.0) {

            if (debug>=5) {
              printf("        sum of inputs > 0.\n");
              fflush(stdout);
            }

            /* state(neuron) = CHARGING_UP  */
            neuron[current_event.object_sub_id].state = (int) CHARGING_UP;

            /* charge_start_time = current_time */
            neuron[current_event.object_sub_id].charge_start_time =
              current_event.time;

            /* charge_start_value = 0 */
            neuron[current_event.object_sub_id].charge_start_value = 0.0;

            /* charge_rate = sum of inputs  */
            neuron[current_event.object_sub_id].charge_rate =
              input_sum;

            /* expected_spike_time = current_time +  */
                                    /* vthresh*capacitance/charge_rate */
            neuron[current_event.object_sub_id].expected_spike_time =
              current_event.time + threshold * CAPACITANCE / input_sum;

            /* expected_zero_time = INFINITY */
            neuron[current_event.object_sub_id].expected_zero_time = INFINITY;

            /* push NEURON_SPIKE event at expected_spike_time */
            push_event(neuron[current_event.object_sub_id].expected_spike_time,
              current_event.object_id,(int)NEURON_SPIKE,
               current_event.object_sub_id);

          /* ELSE */
          } else {

            if (debug>=5) {
              printf("        sum of inputs <= 0.\n");
              fflush(stdout);
            }

            /* state(neuron) = LOW_DISCHARGED */
            neuron[current_event.object_sub_id].state = (int) LOW_DISCHARGED;

            /* charge_start_value = 0 */
            neuron[current_event.object_sub_id].charge_start_value = 0.0;

            /* charge_rate = 0 */
            neuron[current_event.object_sub_id].charge_rate = 0.0;

            /* expected_spike_time = INFINITY */
            neuron[current_event.object_sub_id].expected_spike_time = INFINITY;

            /* expected_zero_time = INFINITY */
            neuron[current_event.object_sub_id].expected_zero_time = INFINITY;

          /* END IF */
          }

          if (debug>=5) dump_object(current_event.object_id);

        /* END CASE (NEURON_REFRACT_OFF): */
        break;


        /* CASE INPUT_SPIKE */
        case INPUT_SPIKE:

          if (debug>=5) {
            printf("      INPUT_SPIKE event.\n");
            fflush(stdout);
          }

          /* write point */
          write_point(current_event.time,current_event.object_id,
                      0.0, FALSE);
          write_point(current_event.time,current_event.object_id,
                      HIGH_VOLTAGE, FALSE);
          write_point(current_event.time,current_event.object_id,
                      0.0, FALSE);

          /* push SYNAPSE_TRANS_ON event for synapses with this input */
          for (i=0; i<synapse_num; i++) {
            if (synapse[i].object_from == current_event.object_id) {
              push_event(current_event.time,synapse[i].id,
                (int)SYNAPSE_TRANS_ON,i);
            }
          }

          /* push BUFFER_SPIKE event for buffers with this input */
          for (i=0; i<buffer_num; i++) {
            if (buffer[i].object_from == current_event.object_id) {
              push_event(current_event.time+buffer[i].t_delay,
                buffer[i].id,(int)BUFFER_SPIKE,i);
            }
          }

          /* push next INPUT_SPIKE event */
          if (fgets(network_spec,MAX_CHARS_PER_LINE,evt_file) != NULL) {
            ptokens[0] = (char *) strtok(network_spec," ,\n");
            numtokens=1;
            while ((ptokens[numtokens]=(char *) strtok((char *)NULL," ,\n"))
                    !=NULL) numtokens++;
            push_event(atof(ptokens[1]), atoi(ptokens[0]),(int)INPUT_SPIKE,
              object_map[atoi(ptokens[0])].sub_id);
          }

        /* END CASE INPUT_SPIKE */
        break;

        /* CASE (NEURON_SPIKE): */
        case NEURON_SPIKE:

          if (debug>=5) {
            printf("      NEURON_SPIKE event.\n");
            fflush(stdout);
          }

          if (debug>=5) dump_object(current_event.object_id);

          /* write point */
          write_point(current_event.time,current_event.object_id,
                      threshold, FALSE);
          write_point(current_event.time,current_event.object_id,
                      HIGH_VOLTAGE, FALSE);
          write_point(current_event.time,current_event.object_id,
                      0.0, FALSE);
          write_point(current_event.time,current_event.object_id,
                      displayed_refractory_voltage, FALSE);

          /* adjust weight of adapting synapses */

          /* push SYNAPSE_TRANS_ON for synapses with this neuron as input */
          for (i=0; i<synapse_num; i++) {
            if (synapse[i].object_from == current_event.object_id) {
              push_event(current_event.time,synapse[i].id,
                (int)SYNAPSE_TRANS_ON,i);
            }
          }

          /* push BUFFER_SPIKE event for buffers with this input */
          for (i=0; i<buffer_num; i++) {
            if (buffer[i].object_from == current_event.object_id) {
              push_event(current_event.time+buffer[i].t_delay,
                buffer[i].id,(int)BUFFER_SPIKE,i);
            }
          }

          /* state(neuron) = LOW_REFRACTORY  */
          neuron[current_event.object_sub_id].state = (int) LOW_REFRACTORY;
          neuron[current_event.object_sub_id].last_spike_time =
            current_event.time;

          /* push NEURON_REFRACT_OFF event at current_time + refract_period */
          push_event(current_event.time +
            neuron[current_event.object_sub_id].t_refract,
            current_event.object_id,(int)NEURON_REFRACT_OFF,
            current_event.object_sub_id);

          if (debug>=5) dump_object(current_event.object_id);

        /* END CASE (NEURON_SPIKE): */
        break;

        /* CASE BUFFER_SPIKE */
        case BUFFER_SPIKE:

          if (debug>=5) {
            printf("      BUFFER_SPIKE event.\n");
            fflush(stdout);
          }

          /* write point */
          write_point(current_event.time,current_event.object_id,
                      0.0, FALSE);
          write_point(current_event.time,current_event.object_id,
                      HIGH_VOLTAGE, FALSE);
          write_point(current_event.time,current_event.object_id,
                      0.0, FALSE);

          /* push SYNAPSE_TRANS_ON event for synapses with this input */
          for (i=0; i<synapse_num; i++) {
            if (synapse[i].object_from == current_event.object_id) {
              push_event(current_event.time,synapse[i].id,
                (int)SYNAPSE_TRANS_ON,i);
            }
          }

          /* push BUFFER_SPIKE event for buffers with this input */
          for (i=0; i<buffer_num; i++) {
            if (strcmp(object_map[buffer[i].id].from_node_label, 
                 object_map[current_event.object_id].to_node_label)==0) {
              push_event(current_event.time+buffer[i].t_delay,
                 buffer[i].id,(int)BUFFER_SPIKE,i);
            }
          }

        /* END CASE BUFFER_SPIKE */
        break;

        /* CASE (NEURON_ZERO): */
        case NEURON_ZERO:

          if (debug>=5) {
            printf("      NEURON_ZERO event.\n");
            fflush(stdout);
          }
          if (debug>=5) dump_object(current_event.object_id);

          /* write event to output file */
          write_point(current_event.time,current_event.object_id,
                      0.0, FALSE);

          /* state(neuron) = LOW_DISCHARGED  */
          neuron[current_event.object_sub_id].state = (int) LOW_DISCHARGED;

          if (debug>=5) dump_object(current_event.object_id);

        /* END CASE (NEURON_ZERO): */
        break;

        /* CASE (SYNAPSE_TRANS_ON): */
        /* CASE (SYNAPSE_TRANS_OFF): */
        case SYNAPSE_TRANS_ON:
        case SYNAPSE_TRANS_OFF:

          
          synapse_value_changed = FALSE;

          /* IF SYNAPSE_TRANS_ON */
          if (current_event.type == (int) SYNAPSE_TRANS_ON) {

            if (debug>=5) {
              printf("      SYNAPSE_TRANS_ON event.\n");
              fflush(stdout);
            }

            if (debug>=5) dump_object(current_event.object_id);

            /* IF state<cliplevel */
            if (synapse[current_event.object_sub_id].state <
                synapse[current_event.object_sub_id].cliplevel) {

              /* write point to output file */
              write_point(current_event.time,current_event.object_id,
                (float) TRANS_ON_HEIGHT*HIGH_VOLTAGE*
                synapse[current_event.object_sub_id].state/
                synapse[current_event.object_sub_id].cliplevel, FALSE);
              write_point(current_event.time,current_event.object_id,
                (float) TRANS_ON_HEIGHT*HIGH_VOLTAGE*
                (synapse[current_event.object_sub_id].state+1)/
                synapse[current_event.object_sub_id].cliplevel, FALSE);

              synapse_value_changed= TRUE;

            /* END IF */
            }

            /* IF cliplevel>1 */
            if (synapse[current_event.object_sub_id].cliplevel > 1) {

              /* increment state */
              synapse[current_event.object_sub_id].state += 1;
 
            /* ELSE */
            } else {
            
              /* state = 1 */
              synapse[current_event.object_sub_id].state = 1;

              /* IF synapse value stayed the same i.e. clipped at 1) */
              if (!synapse_value_changed) {

                /* pop the existing SYNAPSE_TRANS_OFF event */
                pop_event(
                  synapse[current_event.object_sub_id].expected_off_time,
                  current_event.object_id, (int)SYNAPSE_TRANS_OFF);

              /* END IF synapse value stayed the same */
              }
 
              /* compute new expected_off_time */
              synapse[current_event.object_sub_id].expected_off_time = 
                 current_event.time + 
                 synapse[current_event.object_sub_id].t_neuro;

            /* END IF cliplevel>1 */
            }

            /* push SYNAPSE_TRANS_OFF event */
            push_event(current_event.time +
              synapse[current_event.object_sub_id].t_neuro,
              current_event.object_id,(int)SYNAPSE_TRANS_OFF,
              current_event.object_sub_id);

          /* ELSE */
          } else {

            if (debug>=5) {
              printf("      SYNAPSE_TRANS_OFF event.\n");
              fflush(stdout);
            }

            if (debug>=5) dump_object(current_event.object_id);

            /* IF state<=cliplevel */
            if (synapse[current_event.object_sub_id].state <=
                synapse[current_event.object_sub_id].cliplevel) {

              /* write point to output file */
              write_point(current_event.time,current_event.object_id,
                (float) TRANS_ON_HEIGHT*HIGH_VOLTAGE*
                synapse[current_event.object_sub_id].state/
                synapse[current_event.object_sub_id].cliplevel, FALSE);
              write_point(current_event.time,current_event.object_id,
                (float) TRANS_ON_HEIGHT*HIGH_VOLTAGE*
                (synapse[current_event.object_sub_id].state-1)/
                synapse[current_event.object_sub_id].cliplevel, FALSE);

              synapse_value_changed = TRUE;

            /* END IF */
            }

            /* IF cliplevel>1 */
            if (synapse[current_event.object_sub_id].cliplevel > 1) {

              /* decrement state */
              synapse[current_event.object_sub_id].state -= 1;
 
            /* ELSE */
            } else {

              /* expected_off_time = INFINITY */
              synapse[current_event.object_sub_id].expected_off_time 
                 = INFINITY;

              /* state = 0 */
              synapse[current_event.object_sub_id].state = 0; 

            /* END IF cliplevel>1 */
            }

          /* END IF SYNAPSE_TRANS_ON*/
          }

          if (debug>=5) dump_object(current_event.object_id);

          /* +++ */
          /* IF synapse_value_changed */
          if (synapse_value_changed) {
  
            if (debug>=5) {
              printf("      identifying destination neuron.\n");
              fflush(stdout);
            }
            /* identify the index of the destination neuron */
            dest_index = 0;
            while (neuron[dest_index].id !=
                  synapse[current_event.object_sub_id].object_to) dest_index++;
            dest_id = neuron[dest_index].id;
  
            if (debug>=5) dump_object(dest_id);
  
            /* SWITCH (state(neuron)) */
            switch (neuron[dest_index].state) {
  
              /* CASE LOW_DISCHARGED  */
              case LOW_DISCHARGED:
  
                if (debug>=5) {
                  printf("        state of neuron = LOW_DISCHARGED.\n");
                  fflush(stdout);
                }
  
                /* write point to output file for neuron */
                write_point(current_event.time,dest_id, 0.0, FALSE);
  
                /* compute sum of inputs for given neuron */
                /* sum of inputs = tonic + sum (synapse_weight*synapse_state) */
                input_sum = neuron[dest_index].i_tonic;
                for (i=0; i<synapse_num; i++) {
                  if (synapse[i].object_to == dest_id) {
                    if (synapse[i].state != 0) {
                      if (synapse[i].state < synapse[i].cliplevel) {
                        input_sum += synapse[i].i_weight*synapse[i].state;
                      } else {
                        input_sum += synapse[i].i_weight*synapse[i].cliplevel;
                      }
                    }
                  }
                }
  
                /* IF (sum of inputs > 0) */
                if (input_sum > 0.0) {
  
                  if (debug>=5) {
                    printf("          input_sum > 0.0\n");
                    fflush(stdout);
                  }
  
                  /* state(neuron) = CHARGING_UP  */
                  neuron[dest_index].state = (int) CHARGING_UP;
  
                  /* charge_start_time = current_time */
                  neuron[dest_index].charge_start_time = current_event.time;
  
                  /* charge_start_value = 0 */
                  neuron[dest_index].charge_start_value = 0.0;
  
                  /* charge_rate = sum of inputs  */
                  neuron[dest_index].charge_rate = input_sum;
  
                  /* expected_spike_time = current_time +  */
                                          /* vthresh*capacitance/charge_rate */
                  neuron[dest_index].expected_spike_time =
                    current_event.time + threshold * CAPACITANCE / input_sum;
  
                  /* expected_zero_time = INFINITY */
                  neuron[dest_index].expected_zero_time =
                    INFINITY;
  
                  /* push NEURON_SPIKE event at expected_spike_time */
                  push_event( neuron[dest_index].expected_spike_time,
                    dest_id,(int)NEURON_SPIKE,dest_index);
  
                /* END IF (sum of inputs > 0) */
                }
  
              /* END CASE LOW_DISCHARGED  */
              break;
  
              /* CASE CHARGING_UP */
              /* CASE CHARGING_DOWN: */
              /* CASE CHARGED_STEADY: */
              case CHARGING_UP:
              case CHARGING_DOWN:
              case CHARGED_STEADY:
  
                /* IF state(neuron) = CHARGING_UP */
                if (neuron[dest_index].state == (int) CHARGING_UP) {
  
                  if (debug>=5) {
                    printf("        state of neuron = CHARGING_UP.\n");
                    fflush(stdout);
                  }
  
                  /* remove NEURON_SPIKE event at prev. expected_spike_time */
                  pop_event(neuron[dest_index].expected_spike_time, dest_id,
                    (int)NEURON_SPIKE);
  
                /* END IF */
                }
  
                /* IF state(neuron) = CHARGING_DOWN */
                if (neuron[dest_index].state == (int) CHARGING_DOWN) {
  
                  if (debug>=5) {
                    printf("        state of neuron = CHARGING_DOWN.\n");
                    fflush(stdout);
                  }
  
                  /* remove NEURON_ZERO event at previous expected_zero_time */
                  pop_event(neuron[dest_index].expected_zero_time, dest_id,
                    (int)NEURON_ZERO);
  
                /* END IF */
                }
  
                /* IF state(neuron) = CHARGED_STEADY */
                if (neuron[dest_index].state == (int) CHARGED_STEADY) {
  
                  if (debug>=5) {
                    printf("        state of neuron = CHARGED_STEADY.\n");
                    fflush(stdout);
                  }
  
                /* END IF */
                }
  
                /* charge_start_value += */
                /* (current_time - charge_start_time)*charge_rate/capacitance */
                neuron[dest_index].charge_start_value +=
                  (current_event.time - neuron[dest_index].charge_start_time)*
                  neuron[dest_index].charge_rate/CAPACITANCE;
  
                /* write point to output file for neuron */
                write_point(current_event.time,dest_id, 
                      neuron[dest_index].charge_start_value, FALSE);
  
                /* compute sum of inputs for given neuron */
                /* sum of inputs = tonic + sum (synapse_weight*synapse_state) */
                input_sum = neuron[dest_index].i_tonic;
                for (i=0; i<synapse_num; i++) {
                  if (synapse[i].object_to == dest_id) {
                    if (synapse[i].state != 0) {
                      if (synapse[i].state < synapse[i].cliplevel) {
                        input_sum += synapse[i].i_weight*synapse[i].state;
                      } else {
                        input_sum += synapse[i].i_weight*synapse[i].cliplevel;
                      }
                    }
                  }
                }
  
                /* IF (sum of inputs > 0) */
                if (input_sum > 0.0) {
  
                  if (debug>=5) {
                    printf("          input_sum > 0.0\n");
                    fflush(stdout);
                  }
  
                  /* state(neuron) = CHARGING_UP  */
                  neuron[dest_index].state = (int) CHARGING_UP;
  
                  /* charge_start_time = current_time */
                  neuron[dest_index].charge_start_time = current_event.time;
  
                  /* charge_rate = sum of inputs  */
                  neuron[dest_index].charge_rate = input_sum;
  
                  /* expected_spike_time = current_time +  */
                     /* (vthresh-charge_start_value)*capacitance/charge_rate */
                  neuron[dest_index].expected_spike_time =
                    current_event.time +
                    (threshold -
                     neuron[dest_index].charge_start_value) *
                    CAPACITANCE / input_sum;
  
                  /* expected_zero_time = INFINITY */
                  neuron[dest_index].expected_zero_time =
                    INFINITY;
  
                  /* push NEURON_SPIKE event at expected_spike_time */
                  push_event(neuron[dest_index].expected_spike_time,
                    dest_id,(int)NEURON_SPIKE,dest_index);
  
                /* END IF (sum of inputs > 0) */
                }
  
                /* IF (sum of inputs == 0) */
                if (input_sum == 0.0) {
  
                  if (debug>=5) {
                    printf("          input_sum = 0.0\n");
                    fflush(stdout);
                  }
  
                  /* state(neuron) = CHARGED_STEADY  */
                  neuron[dest_index].state = (int) CHARGED_STEADY;
  
                  /* charge_rate = 0 */
                  neuron[dest_index].charge_rate = 0.0;
  
                /* END IF   */
                }
  
                /* IF (sum of inputs < 0) */
                if (input_sum < 0.0) {
  
                  if (debug>=5) {
                    printf("          input_sum = 0.0\n");
                    fflush(stdout);
                  }
  
                  /* state(neuron) = CHARGING_DOWN  */
                  neuron[dest_index].state = (int) CHARGING_DOWN;
  
                  /* charge_start_time = current_time */
                  neuron[dest_index].charge_start_time = current_event.time;
  
                  /* charge_rate = sum of inputs  */
                  neuron[dest_index].charge_rate = input_sum;
  
                  /* expected_zero_time = current_time -  */
                     /* charge_start_value*capacitance/charge_rate */
                  neuron[dest_index].expected_zero_time =
                    current_event.time -
                    neuron[dest_index].charge_start_value *
                    CAPACITANCE / input_sum;
  
                  /* expected_spike_time= INFINITY */
                  neuron[dest_index].expected_spike_time = INFINITY;
  
                  /* push NEURON_ZERO event at expected_zero_time */
                  push_event(neuron[dest_index].expected_zero_time,
                    dest_id,(int)NEURON_ZERO,dest_index);
  
                /* END IF (sum of inputs < 0) */
                }
  
              /* END CASE CHARGING_UP, CHARGING_DOWN, CHARGED_STEADY */
              break;
  
            /* END SWITCH (state(neuron)) */
            }
  
            if (debug>=5) dump_object(dest_id);
  
          /* +++  */
          /* END IF (synapse_value_changed) */
          }

        /* END CASE (SYNAPSE_TRANS_ON, SYNAPSE_TRANS_OFF): */
        break;

        /* CASE (END_RUNTIME): */
        case END_RUNTIME:

          if (debug>=5) {
            printf("      END_RUNTIME event.\n");
            fflush(stdout);
          }

          /* write final point and close output file for each object */
          for (i=0; i<MAX_NUM_OBJECTS; i++) {

            /* SWITCH (object type) */
            switch (object_map[i].type) {

              /* CASE NEURON */
              case NEURON:

                /* SWITCH (state) */
                switch (neuron[object_map[i].sub_id].state) {

                  case LOW_DISCHARGED:
                    write_value = 0.0;
                  break;

                  case LOW_REFRACTORY:
                    write_value = displayed_refractory_voltage +
                      (0.0 - displayed_refractory_voltage)*(current_event.time
                        - neuron[object_map[i].sub_id].last_spike_time)/
                        neuron[object_map[i].sub_id].t_refract;
                  break;

                  case CHARGING_DOWN:
                  case CHARGING_UP:
                    write_value =
                      neuron[object_map[i].sub_id].charge_start_value +
                      neuron[object_map[i].sub_id].charge_rate *
                      (current_event.time -
                        neuron[object_map[i].sub_id].charge_start_time)/
                        CAPACITANCE;
                  break;

                  case CHARGED_STEADY:
                    write_value =
                      neuron[object_map[i].sub_id].charge_start_value;
                  break;

                /* END SWITCH (state) */
                }

                /* write point */
                write_point(current_event.time,i,write_value, TRUE);

              /* END CASE NEURON */
              break;

              /* CASE INPUT */
              case INPUT:

                /* write point */
                write_point(current_event.time,i,0.0, TRUE);

              /* END CASE INPUT */
              break;

              /* CASE BUFFER */
              case BUFFER:

                /* write point */
                write_point(current_event.time,i,0.0, TRUE);

              /* END CASE BUFFER */
              break;


              /* CASE SYNAPSE */
              case SYNAPSE:

                if (synapse[object_map[i].sub_id].state<
                  synapse[object_map[i].sub_id].cliplevel) {
                    write_value = TRANS_ON_HEIGHT*HIGH_VOLTAGE*
                      synapse[object_map[i].sub_id].state/
                      synapse[object_map[i].sub_id].cliplevel;
                } else {
                    write_value = TRANS_ON_HEIGHT*HIGH_VOLTAGE;
                }

                /* write point */
                write_point(current_event.time,i,write_value, TRUE);

              /* END CASE SYNAPSE */
              break;

              case LAST_OBJECT_ENUM:
              break;

            /* END SWITCH (object type) */
            }

          /* END FOR each object */
          }


        /* END CASE (END_RUNTIME): */
        break;

      /* END SWITCH (event.type) */
      }

      if (debug>=5) {
        printf("          about to look for next event.\n");
        printf("          existing event list:\n");
        fflush(stdout);
        dump_events(-1);
      }

    } /* END WHILE (events exist within specified time window) */

    /* print number of events processed */
    printf("\n  number of events processed = %d.\n",event_counter);


    /* concatenate tempobj files into one output file */
    switch(output_format) {
      case (int) CVIEW:
        printf("\n  writing file %s\n",dat_file_name);
        sprintf(command_string,"cat tempobj* > %s\n",dat_file_name);
        break;
      case (int) MATHEMATICA:
        printf("\n  writing file %s\n",math_file_name);
        sprintf(command_string,"cat tempobj* > %s\n",math_file_name);
        break;
      case (int) GNUPLOT:
        printf("\n  writing file %s\n",dat_file_name);
        sprintf(command_string,"cat tempobj* > %s\n",dat_file_name);

        /* open .plt file for writing */
        printf("  writing file %s\n",plt_file_name);
        if ((plt_file = fopen(plt_file_name,"w+"))==NULL) {
          printf("\n  main:  unable to open file:  %s\n",plt_file_name);
          printf("  Exiting.\n");
          fflush(stdout); exit(1);
        }

        fprintf(plt_file,"set nokey\n");
        fprintf(plt_file,"set yrange [%f:1.5]\n",-saved_file_num+0.5);
        fprintf(plt_file,"set xlabel \"Time (ms)\"\n");
        fprintf(plt_file,"set title \"%s\"\n",dat_file_name);
        fflush(plt_file);

        /* write ytic label list */
        for (i=0; i<saved_file_num; i++) {
          jj=0;
          while (i!=object_map[jj].display_order) jj++;
 
          /* if label name is very long */
          if (strlen(object_map[jj].label)>(int)max_mathlabel_length) {
       
            /* make abbreviated name out of first + last few characters */
            source = &object_map[jj].label[0];
            dest = &dummy_label[0];
            for (kk=0;kk<(int)(max_mathlabel_length-1)/2;kk++) {
              *dest++ = *source++;
            }
            *dest++ = '*';
            for (kk=0;kk<strlen(object_map[jj].label)-
                max_mathlabel_length+1;kk++) {
              source++;
            }
            for (kk=0;kk< (max_mathlabel_length)-1-
                (int)(max_mathlabel_length-1)/2;kk++) {
              *dest++ = *source++;
            }
            *dest++ = '\0';
          } else {
            /* just take name as is */
            strcpy(dummy_label,object_map[jj].label);
          }
        
          if (i==0) {
            fprintf(plt_file,"set ytics (\"%s\" 0", dummy_label); 
            fflush(plt_file);
          } else {
            fprintf(plt_file, ", \"%s\" %d", dummy_label,-i); 
            fflush(plt_file);
          }
        }
        fprintf(plt_file,")\n");
        fflush(plt_file);

        fprintf(plt_file,"plot \"%s\" with lines\n",dat_file_name);
        fprintf(plt_file,"pause -1\n");
        fflush(plt_file);

        fclose(plt_file);
        
        break;

      case (int) XVGR: {
        /* open .plt file for writing */
        if ((plt_file = fopen(plt_file_name,"w+"))==NULL) {
          printf("\n  main:  unable to open file:  %s\n",plt_file_name);
          printf("  Exiting.\n");
          fflush(stdout); exit(1);
        }

	fprintf(plt_file, "@ world xmax %d\n@ world xmin 0\n", (int)runtime);
	fprintf(plt_file, "@ world ymax 1.5\n@ world ymin %d\n", 
                -saved_file_num-.5);

        fprintf(plt_file,"@ title \"%s\"\n",vgr_file_name);

	fprintf(plt_file, "@ xaxis label \"Time (ms)\"\n@ xaxis ticklabel on\n");
	fprintf(plt_file, "@ xaxis ticklabel type auto\n@ xaxis ticklabel prec 0\n");
	fprintf(plt_file, "@ xaxis tick on\n@ xaxis tick out\n@ xaxis tick op bottom\n");
	fprintf(plt_file, "@ xaxis tick type auto\n@ xaxis tick spec 0\n");
	fprintf(plt_file, "@ xaxis tick major %d\n@ xaxis tick minor %f\n", (int)runtime/5, 0.2*(runtime/5));

        fprintf(plt_file,"@ xaxis label \"Time (ms)\"\n");
        fprintf(plt_file,"@ xaxis tick type auto\n@ xaxis tick spec 0\n");

        fprintf(plt_file,"@ yaxis ticklabel on\n@ yaxis ticklabel type spec\n");
        fprintf(plt_file,"@ yaxis tick off\n@ yaxis tick type spec\n@ yaxis tick spec %d\n",saved_file_num);

	fflush(plt_file);

        /* write ytic label list */
        for (i=0; i<saved_file_num; i++) {
          jj=0;
          while (i!=object_map[jj].display_order) jj++;

          /* if label name is very long */
          if (strlen(object_map[jj].label)>(int)max_mathlabel_length) {
 
            /* make abbreviated name out of first + last few characters */
            source = &object_map[jj].label[0];
            dest = &dummy_label[0];
            for (kk=0;kk<(int)(max_mathlabel_length-1)/2;kk++) {
                    *dest++ = *source++;
            }
            *dest++ = '*';
            for (kk=0;kk<strlen(object_map[jj].label)-
                max_mathlabel_length+1;kk++) {
              source++;
            }
            for (kk=0;kk< (max_mathlabel_length)-1-
                (int)(max_mathlabel_length-1)/2;kk++) {
              *dest++ = *source++;
            }
            *dest++ = '\0';
          } else {
            /* just take name as is */
            strcpy(dummy_label,object_map[jj].label);
          }

          fprintf(plt_file,"@ yaxis tick %d,%d\n",i,-i); 
          fprintf(plt_file,"@ yaxis ticklabel %d,\"%s\"\n",
                i,dummy_label); 
          fflush(plt_file);
        }

	fprintf(plt_file, "@WITH G0\n");
	fflush(plt_file);

        printf("\n  writing file %s\n",vgr_file_name);
        sprintf(command_string,"cat %s tempobj* > %s\n",
            plt_file_name,vgr_file_name);
        break;
      }
    }

    fflush(stdout);
    system(command_string);

    /* remove files */
    if (debug<10) {
      system("rm tempobj*\n");

      sprintf(command_string,"rm %s\n",evt_file_name);
      system(command_string);
    }

    /* print goodbye message */
    printf("  done.\n\n\n"); fflush(stdout);

/* END */
}

/*****************************************************************/
void
dump_object(object_id)
/*****************************************************************/
int object_id;
{
    extern struct neuron_struct neuron[];
    extern struct synapse_struct synapse[];
    extern struct buffer_struct buffer[];
    extern struct object_map_struct object_map[];
    extern char *neuron_state_string[];
    extern char *synapse_state_string[];
    extern char *object_type_string[];
    int sub_id;


    printf("\n\n  STATE OF OBJECT   %d: \n",object_id);
    printf("    object_type:     %s\n",
      object_type_string[object_map[object_id].type]);
    printf("    object_sub_id:   %d\n", object_map[object_id].sub_id);
    printf("    object_label:    %s\n", object_map[object_id].label);
    printf("    from_node:       %s\n", object_map[object_id].from_node_label);
    printf("    to_node:         %s\n", object_map[object_id].to_node_label);
    printf("    display_order:   %d\n", object_map[object_id].display_order);
    printf("    saved_file_num:  %d\n", object_map[object_id].saved_file_num);

    sub_id = object_map[object_id].sub_id;

    /* IF NEURON */
    if (object_map[object_id].type == (int) NEURON) {

      printf("      t_refract:            %f\n",neuron[sub_id].t_refract);
      printf("      i_tonic:              %f\n", neuron[sub_id].i_tonic);
      printf("      state:                %s\n", 
        neuron_state_string[neuron[sub_id].state]);
      printf("      charge_rate:          %f\n", neuron[sub_id].charge_rate);
      printf("      charge_start_time:    %f\n",
        neuron[sub_id].charge_start_time);
      printf("      charge_start_value:   %f\n",
        neuron[sub_id].charge_start_value);
      printf("      expected_spike_time:  %f\n",
        neuron[sub_id].expected_spike_time);
      printf("      expected_zero_time:   %f\n",
        neuron[sub_id].expected_zero_time);
      printf("      last_spike_time:      %f\n", 
        neuron[sub_id].last_spike_time);

    /* END IF NEURON */
    }

    /* IF BUFFER */
    if (object_map[object_id].type == (int) BUFFER) {

      printf("      object_from:        %d\n",buffer[sub_id].object_from);
      printf("      t_delay:            %f\n",buffer[sub_id].t_delay);

    /* END IF BUFFER */
    }

    /* IF SYNAPSE */
    if (object_map[object_id].type == (int) SYNAPSE) {

      printf("      object_from:        %d\n",synapse[sub_id].object_from);
      printf("      object_to:          %d\n",synapse[sub_id].object_to);
      printf("      i_weight:           %f\n",synapse[sub_id].i_weight);
      printf("      t_neuro:            %f\n",synapse[sub_id].t_neuro);
      printf("      state:              %d\n",synapse[sub_id].state);
      printf("      cliplevel:          %d\n",synapse[sub_id].cliplevel);

    /* END IF SYNAPSE */
    }

    printf("\n"); fflush(stdout);
}
/*****************************************************************/
void
dump_map()
/*****************************************************************/
{
    extern struct object_map_struct object_map[];
    extern char *object_type_string[];
    int i;

    printf("\n\n                OBJECT MAP                   \n\n");
    printf("                 sub  display           \n");
    printf("    id   type     id   order    label   \n");
    printf("  --------------------------------------\n");
    for (i=0; i<MAX_NUM_OBJECTS; i++) {
      if (object_map[i].type != (int) LAST_OBJECT_ENUM) {
        if (object_map[i].display_order != NO_DISPLAY_VALUE) {
          printf("  %4d  %-8s %4d %4d   %s\n", 
               i,object_type_string[object_map[i].type],
               object_map[i].sub_id, object_map[i].display_order,
               object_map[i].label);
        } else {
          printf("  %4d  %-8s %4d        %s\n", 
               i, object_type_string[object_map[i].type],
    object_map[i].sub_id, object_map[i].label);
        } 
      }
    }
    printf("\n\n");
}


/*****************************************************************/
void
dump_events(next_event)
/*****************************************************************/
int next_event;
{
    extern struct event_struct event[];
    extern float runtime;
    extern int max_event_index;
    extern char *event_type_string[];
    int i;

    printf("\n  number    time(ms)   event_type        object_id  object_sub_id\n");
    printf("  ---------------------------------------------------------------\n"
);
    for (i=0; i<=max_event_index; i++) {
      if (event[i].time <= runtime) {
        if (i == next_event) {
          printf("  %4d     %7.3f    %-18s %5d  %10d  <**NEXT**> \n",
            i,event[i].time, event_type_string[event[i].type],
            event[i].object_id, event[i].object_sub_id);
        } else {
          printf("  %4d     %7.3f    %-18s %5d  %10d\n",
            i,event[i].time, event_type_string[event[i].type],
            event[i].object_id, event[i].object_sub_id);
        }
      }
    }

    printf("\n   max_event_index = %d\n\n",max_event_index); fflush(stdout);
}
/*****************************************************************/
void
pop_specified_event(event_num)
/*****************************************************************/
int event_num;
{
    extern struct event_struct event[];
    extern char *event_type_string[];
    extern int max_event_index;
    extern int debug;

    /* pop the event */
    if (debug>=5) {
      printf("\n\n  popping event:\n");
      printf("    location: %d\n",event_num);
      printf("    time: %f\n",event[event_num].time);
      printf("    object_id: %d\n",event[event_num].object_id);
      printf("    event_type: %s\n",event_type_string[event[event_num].type]);
      printf("    object_sub_id: %d\n",event[event_num].object_sub_id);
      fflush(stdout);
    }
    event[event_num].time = INFINITY;
    event[event_num].object_id = -1;
    event[event_num].type = (int) LAST_EVENT_ENUM;
    event[event_num].object_sub_id = -1;

    /* adjust max_event_index downwards to next real event */
    while (event[max_event_index].type==(int)LAST_EVENT_ENUM) max_event_index--;

}
/*****************************************************************/
void
pop_event(time, object_id, event_type)
/*****************************************************************/
float time;
int object_id;
int event_type;
{
    extern struct event_struct event[];
    extern char *event_type_string[];
    extern int max_event_index;
    extern int debug;
    int i;

    i=0;
    /* find the event */
    while ((event[i].time != time || event[i].object_id != object_id ||
           event[i].type != event_type)&& i<=max_event_index) i++;

    /* pop it */
    if (i<=max_event_index) {
      if (debug>=5) {
        printf("\n\n  popping event:\n");
        printf("    location: %d\n",i);
        printf("    time: %f\n",event[i].time);
        printf("    object_id: %d\n",event[i].object_id);
        printf("    event_type: %s\n",event_type_string[event[i].type]);
        printf("    object_sub_id: %d\n",event[i].object_sub_id);
        fflush(stdout);
      }
      event[i].time = INFINITY;
      event[i].object_id = -1;
      event[i].type = (int) LAST_EVENT_ENUM;
      event[i].object_sub_id = -1;
    } else {
      i = -1;
      printf ("  pop_event:  event not found\n");
      printf ("      time:  %f\n",time);
      printf ("      object_id:  %d\n",object_id);
      printf ("      event_type:  %s\n",event_type_string[event_type]);
      printf("  Exiting.\n");
      fflush(stdout);
      exit(1);
    }

    /* adjust max_event_index downwards to next real event */
    while (event[max_event_index].type==(int)LAST_EVENT_ENUM) max_event_index--;
}
/*****************************************************************/
void
push_event(time, object_id, event_type, object_sub_id)
/*****************************************************************/
float time;
int object_id;
int event_type;
int object_sub_id;
{
    extern struct event_struct event[];
    extern int max_event_index;
    extern int debug;
    extern char *event_type_string[];
    int i;

    /* find a free location for the event */
    i = 0;
    while (event[i].type != (int)LAST_EVENT_ENUM && i<MAX_NUM_EVENTS) i++;

    /* handle error condition */
    if (i>=MAX_NUM_EVENTS) {
      printf("  push_event:  no free event locations found\n");
      printf("  Exiting.\n");
      fflush(stdout);
      exit(1);
    }

    if (debug>=5) {
      printf("\n\n  pushing event:\n");
      printf("    location: %d\n",i);
      printf("    time: %f\n",time);
      printf("    object_id: %d\n",object_id);
      printf("    event_type: %s\n",event_type_string[event_type]);
      printf("    object_sub_id: %d\n",object_sub_id);
      fflush(stdout);
    }

    /* push the event */
    event[i].time = time;
    event[i].object_id = object_id;
    event[i].type = event_type;
    event[i].object_sub_id = object_sub_id;

    /* update max_event_index */
    if (i>max_event_index) max_event_index = i;

}
/*****************************************************************/
int
find_next_event()
/*****************************************************************/
{
    extern struct event_struct event[];
    extern float runtime;
    extern int max_event_index;
    extern int debug;
    float next_time;
    float dummy_time;
    int next_index;
    int i;

    next_index = -1;
    next_time = INFINITY;
    for (i=0; i<=max_event_index; i++) {
      dummy_time = event[i].time;
      if (dummy_time < next_time && dummy_time <= runtime) {
        next_index = i;
        next_time = dummy_time;
      }
    }

    if (debug>=5) {
      printf("\n          next event = %d\n",next_index);
      fflush(stdout);
    }

    return next_index;

}

/*****************************************************************/
float
shift_plot_point(value, shift_amount, offset)
/*****************************************************************/
float value;
int shift_amount;
int offset;
{
    float return_value;

    switch (offset) {

      case NONE:
        return_value = value;
      break;

      case UP:
        return_value = (float) shift_amount + 
                       (1.0 - PLOT_MARGIN) * value/HIGH_VOLTAGE;
      break;

      case DOWN:
        return_value = (float) -shift_amount + 
                       (1.0 - PLOT_MARGIN) * value/HIGH_VOLTAGE;
      break;
    }

    return return_value;
}
/*****************************************************************/
void
open_output_files()
/*****************************************************************/
{
    extern struct object_map_struct object_map[];
    extern int output_format;
    extern FILE *object_file[];
    extern int offset;
    extern int saved_file_num;
    extern int object_num;
    int i,remainder,j,jj,kk;
    char suffix_string[SUFFIX_STRING_DIGITS1];
    char file_name[MAX_FILE_NAME_LENGTH];
    char curve_name[MAX_FILE_NAME_LENGTH];
    char dummy_label[MAX_CHARS_PER_LINE];
    char *source;
    char *dest;

    /* open data file for each object with appropriate header */
    saved_file_num = 0;
    for (i=0; i<object_num; i++) {

      if (object_map[i].display_order != NO_DISPLAY_VALUE) { 

        /* generate suffix digits */
        remainder = saved_file_num;
        for (j=0; j<SUFFIX_STRING_DIGITS; j++) {
          suffix_string[j] =
            (int) remainder / (int) pow(10.0,(double)(SUFFIX_STRING_DIGITS - 1 -j));
          remainder -= (int) suffix_string[j] *
            (int) pow(10.0,(double)(SUFFIX_STRING_DIGITS -1 -j));
          suffix_string[j] += '0';
        }
        suffix_string[SUFFIX_STRING_DIGITS] = '\0';
  
        /* generate file_name */
        strcpy(file_name,"tempobj");
        strcat(file_name,suffix_string);
  
        strcpy(curve_name,object_map[i].label);
        object_map[i].saved_file_num = saved_file_num;
  
        /* open the file */
        if ((object_file[saved_file_num] = fopen(file_name,"w"))==NULL) {
          printf("\n  open_output_files: unable to open file:  %s\n",file_name);
          printf("  Exiting.\n"); fflush(stdout); exit(1);
        }
  
        switch(output_format) {
          case (int)CVIEW:

            /* write view header in first file */
            if (saved_file_num==0) {
              fprintf(object_file[0],"1\n");
              fprintf(object_file[0],"# Written from Spike\n");
            }
            fprintf(object_file[saved_file_num],"\npairs\n");
            fprintf(object_file[saved_file_num],"%s\n",curve_name);
            fprintf(object_file[saved_file_num],"Time (ms)\n*\n");
  
            /* write first point */
            fprintf(object_file[saved_file_num],"0.0 %f\n",
              shift_plot_point(0.0,object_map[i].display_order,offset));
            break;

          case (int)MATHEMATICA:

            /* mathematica format */
            /* write header in first file */
            if (saved_file_num==0) {
              fprintf(object_file[0],"(* Mathematica file,");
              fprintf(object_file[0]," written by Spike *)\n\n");
            }

            fprintf(object_file[saved_file_num],"%s = ",curve_name);
  
            /* write first point */
            fprintf(object_file[saved_file_num],"{{0.0, %f}",
              shift_plot_point(0.0,object_map[i].display_order,offset));
            break;

          case (int)GNUPLOT:

            /* write gnuplot header in first file */
            if (saved_file_num==0) {
              fprintf(object_file[0],"# Gnuplot file, written by Spike\n");
            } else {
              fprintf(object_file[saved_file_num],"\n");
            }
  
            /* write first point */
            fprintf(object_file[saved_file_num],"0.0 %f\n",
              shift_plot_point(0.0,object_map[i].display_order,offset));
            break;

          case (int)XVGR:

            /* write XVGR header in first file */
            if (saved_file_num!=0) {
              fprintf(object_file[saved_file_num],"\n");
            }
  
            /* write first point */
            fprintf(object_file[saved_file_num],"0.0 %f\n",
              shift_plot_point(0.0,object_map[i].display_order,offset));
            break;


        }
        saved_file_num++;
      }
    }
}
/*****************************************************************/
void
write_point(time, object_id, voltage, last_point)
/*****************************************************************/
float time;
int object_id;
float voltage;
int last_point;
{
    extern char basic_file_name[MAX_FILE_NAME_LENGTH];
    extern struct object_map_struct object_map[];
    extern int output_format;
    extern FILE *object_file[];
    extern int offset;
    extern int saved_file_num;
    extern int debug;
    char dummy_label[MAX_CHARS_PER_LINE];
    int jj,i,kk;
    char * source;
    char * dest;

    /* IF we want to display this object */
    if ( object_map[object_id].display_order 
         != NO_DISPLAY_VALUE) {

      /* SWITCH (output_format) */
      switch (output_format) {
        case (int)CVIEW:

          fprintf(object_file[object_map[object_id].saved_file_num],"%f %f\n",
            time, shift_plot_point(voltage,object_map[object_id].display_order, 
            offset));
          if (last_point) {
            /* close output file */
            fclose(object_file[object_map[object_id].saved_file_num]);
          }
          break;

        case (int)MATHEMATICA:

          if (!last_point) {
            fprintf(object_file[object_map[object_id].saved_file_num],
              ",\n        {%f, %f}", time, 
              shift_plot_point(voltage,object_map[object_id].display_order,
                             offset));
          } else {
            fprintf(object_file[object_map[object_id].saved_file_num],
              ",\n        {%f, %f}}\n\n", time, 
              shift_plot_point(voltage,object_map[object_id].display_order,
                             offset));

            /* IF last file */
            if (object_map[object_id].saved_file_num == saved_file_num -1) {

              /* write allcurves list */
              jj=0;
              for (i=0; i<object_num; i++) {
                if (object_map[i].display_order != NO_DISPLAY_VALUE) { 

                  if (jj==0) {
                    fprintf(object_file[saved_file_num-1],"allcurves = {%s",
                       object_map[i].label); 
                    fflush(object_file[saved_file_num-1]);
                  } else {
                    fprintf(object_file[saved_file_num-1],", %s",
                         object_map[i].label); 
                    fflush(object_file[saved_file_num-1]);
                  }
                  jj++;
                }
              }
              fprintf(object_file[saved_file_num-1],"}\n\n",
                 object_map[i].label); 
              fflush(object_file[saved_file_num-1]);

              /* write label list */
              for (i=0; i<saved_file_num; i++) {
                jj=0;
                while (i!=object_map[jj].display_order) jj++;
  
                /* if label name is very long */
                if (strlen(object_map[jj].label)>(int)max_mathlabel_length) {
             
                  /* make abbreviated name out of first + last few characters */
                  source = &object_map[jj].label[0];
                  dest = &dummy_label[0];
                  for (kk=0;kk<(int)(max_mathlabel_length-1)/2;kk++) {
                    *dest++ = *source++;
                  }
                  *dest++ = '*';
                  for (kk=0;kk<strlen(object_map[jj].label)-
                      max_mathlabel_length+1;kk++) {
                    source++;
                  }
                  for (kk=0;kk< (max_mathlabel_length)-1-
                      (int)(max_mathlabel_length-1)/2;kk++) {
                    *dest++ = *source++;
                  }
                  *dest++ = '\0';
                } else {
                  /* just take name as is */
                  strcpy(dummy_label,object_map[jj].label);
                }
              
                if (i==0) {
                  fprintf(object_file[saved_file_num-1],
                     "labellist = {{%d,\"%s\"}",
                     -i,dummy_label); 
                  fflush(object_file[saved_file_num-1]);
                } else {
                  fprintf(object_file[saved_file_num-1],
                      ", {%d,\"%s\"}",
                      -i,dummy_label); 
                  fflush(object_file[saved_file_num-1]);
                }
              }
              fprintf(object_file[saved_file_num-1],"}\n\n");
              fflush(object_file[saved_file_num-1]);

              /* write plot command */
              fprintf(object_file[saved_file_num-1],
                "(* Plot the traces *)\n");
              fprintf(object_file[saved_file_num-1],
                "allplot = Show[Table[ListPlot[allcurves[[i]],\n");
              fprintf(object_file[saved_file_num-1],
                "     PlotJoined->True,PlotStyle->Thickness[0.002],\n");
              fprintf(object_file[saved_file_num-1],
                "     DisplayFunction->Identity],{i,Length[allcurves]}],\n");
              fprintf(object_file[saved_file_num-1],
                "     DisplayFunction->$DisplayFunction,\n");
              fprintf(object_file[saved_file_num-1],
                "     Axes->False,Frame->True,PlotRange->All,\n");
              fprintf(object_file[saved_file_num-1],
                "     FrameLabel -> {\"Time (ms)\",\"   \",\" \",\"   \"},\n");
              fprintf(object_file[saved_file_num-1],
                "     FrameTicks->{Automatic,labellist},\n");
              fprintf(object_file[saved_file_num-1],
                "     DisplayFunction->$DisplayFunction,\n");
              fprintf(object_file[saved_file_num-1],
                "     DefaultFont->{\"Helvetica\",10}];\n\n");

              fprintf(object_file[saved_file_num-1],
                "(* Write PostScript to file %s.ps *)\n",basic_file_name);
              fprintf(object_file[saved_file_num-1],
                "Display[\"!psfix -lmarg 1 -tmarg 1 -rmarg 1 -bmarg 2\n");
              fprintf(object_file[saved_file_num-1],
                "     -width 8.5 -height 11> %s.ps\", allplot];\n\n",
                basic_file_name);

            /* END IF last file */
            }

            /* close output file */
            fclose(object_file[object_map[object_id].saved_file_num]);

          /* END IF last point */
          }
          break;

        case (int)GNUPLOT:
        case (int)XVGR:

          fprintf(object_file[object_map[object_id].saved_file_num],"%f %f\n",
            time, shift_plot_point(voltage,object_map[object_id].display_order, 
            offset));
          if (last_point) {
            /* close output file */
            fclose(object_file[object_map[object_id].saved_file_num]);
          }
          break;
         
      /* END SWITCH (output_format) */
      }

    /* END IF we want to display this object */
    }

}

/*****************************************************************/
void
interpret_ntk_file()
/*****************************************************************/
{

    extern struct object_map_struct object_map[];
    extern struct neuron_struct neuron[];
    extern struct synapse_struct synapse[];
    extern struct tonic_struct tonic[];
    extern struct buffer_struct buffer[];
    extern int          neuron_num,synapse_num,input_num,object_num;
    extern int          debug, buffer_num;
    extern char        *object_type_string[];
    extern FILE        *ntk_file;
    extern char         ntk_file_name[];
    extern int          ntk_file_found;
    extern int          tonic_num;
    extern int          num_errors_found;


    FILE               *fopen();
    FILE               *ntk2_file;
    FILE               *ntk3_file;
    FILE               *ntk4_file;
    char                network_spec[MAX_CHARS_PER_LINE];
    char                network_spec_copy[MAX_CHARS_PER_LINE];
    char                buffer_inputs[MAX_NUM_INPUTS][MAX_CHARS_PER_LINE];
    char                buffer_neurons[MAX_NUM_NEURONS][MAX_CHARS_PER_LINE];
    char                buffer_synapses[MAX_NUM_SYNAPSES][MAX_CHARS_PER_LINE];
    char                buffer_buffers[MAX_NUM_BUFFERS][MAX_CHARS_PER_LINE];
    char                from_node_string[MAX_CHARS_PER_LINE];
    char                to_node_string[MAX_CHARS_PER_LINE];
    char               *ptokens[MAX_TOKENS_PER_LINE];
    char                suffix_string[SUFFIX_STRING_DIGITS1];
    char                instance_name[MAX_CHARS_PER_LINE];
    char                cell_type[MAX_CHARS_PER_LINE];
    char                expand_cell_parameters[MAX_NUM_EXPAND_CELL_PARAMETERS][MAX_CHARS_PER_LINE];
    char                instance_parameters[MAX_NUM_EXPAND_CELL_PARAMETERS][MAX_CHARS_PER_LINE];
    int                 ii,i,j,numtokens;
    int                 data_found,end_data_found;
    int                 num_inputs;
    int                 num_neurons;
    int                 num_synapses;
    int                 num_buffers;
    int                 from_node;
    int                 to_node;
    int                 best_index;
    int                 ntk_read_state;
    int                 num_leafs;
    int                 num_parameters;
    int                 found_index;
    int                 duplicate_found;

    /* initialize basic values */
    data_found = FALSE;
    end_data_found = FALSE;

    /* open ntk file */
    if ((ntk_file = fopen(ntk_file_name,"r"))==NULL) {
      /* printf("\n  interpret_ntk_file:  unable to open file:  %s\n",ntk_file_name */
/* ); */
      /* printf("  Exiting.\n"); */
      /* fflush(stdout); exit(1); */
      return;
    }

    ntk_file_found = TRUE;

    printf("  interpreting %s\n",ntk_file_name); fflush(stdout);

    /* open ntk2 file */
    if ((ntk2_file = fopen("tempntk2","w+"))==NULL) {
      printf("\n  interpret_ntk_file:  unable to open file:  tempntk2\n");
      printf("  Exiting.\n");
      fflush(stdout); exit(1);
    }

    num_leafs = 0;
    ntk_read_state = (int)LOOKING;

    /* WHILE more lines exist */
    while (fgets(network_spec,MAX_CHARS_PER_LINE,ntk_file) != NULL) {

      if (debug>=10) {
        printf("    %s",network_spec); fflush(stdout);
      }

      /* copy ntk line over to ntk2 file if it begins with 'c','h', or '.' */
      switch (network_spec[0]) {

        case 'c':
          /* remember the line */
          strcpy(network_spec_copy,network_spec);
          ntk_read_state = (int)GOT_C;
          break;
 
        case '.':
 
          switch (ntk_read_state) {
            case GOT_CH:
            case GOT_H:
              fprintf(ntk2_file,"%s",network_spec); fflush(ntk2_file);
              break;
          }
          ntk_read_state = (int)LOOKING;
          break;

        case 'h':
          switch (ntk_read_state) {

            case LOOKING:
              if (debug>=10) {
                printf("    \nfound top-level cell:\n\n"); fflush(stdout);
              }
              fprintf(ntk2_file,"%s",network_spec); fflush(ntk2_file);
              ntk_read_state = (int)GOT_H;

              /* parse, compare tokens 2 and 3 */

              /* read a line */
              strcpy(network_spec_copy,network_spec);

              /* parse into tokens */
              ptokens[0] = (char *) strtok(network_spec," ,\n");

              numtokens=1;
              while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n"))
                       != NULL) numtokens++;
      
              /* IF tokens 2 and 3 agree exactly */
              if (strcmp(ptokens[1],ptokens[2])==0) {
      
                printf("\n  interpret_ntk_file:  top-level cell is a leaf (remove label from .lgf\n");
                printf("                       and re-run :logntk).\n");
                printf("  Exiting.\n");
                fflush(stdout); exit(1);

              /* END IF tokens 2 and 3 agree exactly */
              }
              break;

            case GOT_C:
              if (debug>=10) {
                printf("    \nfound leaf cell:\n\n"); fflush(stdout);
              }
              num_leafs++;
              fprintf(ntk2_file,"%s",network_spec_copy); 
              fprintf(ntk2_file,"%s",network_spec); fflush(ntk2_file);
              ntk_read_state = (int)GOT_CH; 
              break;

            case GOT_CH:
              fprintf(ntk2_file,"%s",network_spec); fflush(ntk2_file);
              break;

            case GOT_H:
              fprintf(ntk2_file,"%s",network_spec); fflush(ntk2_file);
              break;

          }
          break;
      }

    } /* END WHILE more lines exist */

    if (debug>=1 && num_leafs>0) {
      printf("  flattening hierarchy\n"); 
      fflush(stdout);
    }

    if (debug>=5) {
      printf("\n  number of leaf cells found = %d.\n\n",num_leafs); 
      fflush(stdout);
    }

    /* go back to beginning of ntk2 file */
    fclose(ntk2_file);

    /* next pass throught ntk2 file */
    /* WHILE more leaf cells remain */
    while (num_leafs>0) {

      if ((ntk2_file = fopen("tempntk2","r"))==NULL) {
        printf("\n  interpret_ntk_file:  unable to open file:  tempntk2\n");
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }

      if ((ntk3_file = fopen("tempntk3","w+"))==NULL) {
        printf("\n  interpret_ntk_file:  unable to open file:  tempntk3\n");
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }

      if ((ntk4_file = fopen("tempntk4","w+"))==NULL) {
        printf("\n  interpret_ntk_file:  unable to open file:  tempntk4\n");
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }

      /* read first line of ntk2 file */
      fgets(network_spec,MAX_CHARS_PER_LINE,ntk2_file); 
      
      /* parse the line  */
      ptokens[0] = (char *) strtok(network_spec," ,\n");
  
      numtokens=1;
      while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n")) 
               != NULL) numtokens++;
  
      strcpy(cell_type,ptokens[1]);
      if (debug>=10) {
        printf("   expanding cell type:  %s\n",cell_type); 
        fflush(stdout);
      }

      num_parameters = 0;
      for (i=2; i<numtokens; i++) {
        if (strcmp(ptokens[i],"VDD") != 0 &
            strcmp(ptokens[i],"GND") != 0 &
            strcmp(ptokens[i],";") != 0) {
          strcpy(expand_cell_parameters[num_parameters],ptokens[i]);
          if (debug>=10) {
            printf("          parameters found:  %s\n",
               expand_cell_parameters[num_parameters]); 
            fflush(stdout);
          }
          num_parameters++;
        }
      }

      /* copy following h-lines to ntk3 until next . */
      /* WHILE more lines exist AND first char <> '.' */
      while (fgets(network_spec,MAX_CHARS_PER_LINE,ntk2_file) != NULL
              && network_spec[0] != '.') {

        if (debug>=10) {
          printf("    %s",network_spec); fflush(stdout);
        }

        /* copy ntk line over to ntk3 file */
        fprintf(ntk3_file,"%s",network_spec); fflush(ntk3_file);

      } /* END WHILE more lines exist */

      /* WHILE more lines exist */
      while (fgets(network_spec,MAX_CHARS_PER_LINE,ntk2_file) != NULL) {

        /* IF first character is h */
        if (network_spec[0] == 'h') {

          /* parse the line */
          strcpy(network_spec_copy,network_spec);
          ptokens[0] = (char *) strtok(network_spec," ,\n");
  
          numtokens=1;
          while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n")) 
                   != NULL) numtokens++;

          /* IF this line is an instance of the cell to be expanded */
          if (strcmp(ptokens[1],cell_type)==0) {

            /* expand it */
            /* store instance name and parameters */
            strcpy(instance_name,ptokens[2]);
            for (i=0;i<num_parameters;i++) {
              strcpy(instance_parameters[i],ptokens[i+3]);
            }
     
            /* go back to beginning of ntk3 file */
            fseek(ntk3_file, (long) 0, 0);
 
            /* WHILE more lines in ntk3 file */
            while (fgets(network_spec,MAX_CHARS_PER_LINE,ntk3_file) 
                    != NULL) {

              /* parse the line */
              ptokens[0] = (char *) strtok(network_spec," ,\n");
              numtokens=1;
              while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n"))
                       != NULL) numtokens++;

              /* write out first three words */
              fprintf(ntk4_file,"%s %s %s%s ",ptokens[0],ptokens[1],
                instance_name,ptokens[2]); fflush(ntk4_file);
         
              /* FOR each parameter */
              for (i=3; i<numtokens-3; i++) {

                /* IF this parameter is an object label (i.e. n1, b1, or in1) */
                if ((strcmp(ptokens[1],"NEURON")==0 && i==5) ||
                    (strcmp(ptokens[1],"BUFFER")==0 && i==5) ||
                    (strcmp(ptokens[1],"INPUT")==0 && i==4)) {

                  /* prepend instance_name and write it */
                  fprintf(ntk4_file,"%s%s ",instance_name,ptokens[i]);
                  fflush(ntk4_file);
                  
                /* ELSE */
                } else {

                  /* IF numerical value (first char is digit . + -) */
                  if (*ptokens[i]=='.' || isdigit((int)*ptokens[i]) || 
                      *ptokens[i]=='-' || *ptokens[i]=='+') {

                    /* write it as is */
                    fprintf(ntk4_file,"%s ",ptokens[i]);
                    fflush(ntk4_file);

                  /* ELSE */
                  } else {

                    /* look for this parameter in the instance parameter list */
                    found_index=-1;
                    for (j=0;j<num_parameters;j++) {
                      if (strcmp(ptokens[i],expand_cell_parameters[j]) == 0) 
                        found_index = j;
                    }
 
                    /* IF parameter found */
                    if (found_index != -1) {

                      /* write out replacement parameter */
                      fprintf(ntk4_file,"%s ",instance_parameters[found_index]);
                      fflush(ntk4_file);

                    /* ELSE */
                    } else {

                      /* prepend instance_name and write it */
                      fprintf(ntk4_file,"%s%s ",instance_name,ptokens[i]);
                      fflush(ntk4_file);

                    /* END IF parameter found */
                    }
                  /* END IF numerical value */
                  }
                /* END IF parameter is object label */
                }
              } /* END FOR each parameter */

              /* write out last three parameters (VDD, GND, ;) */
              fprintf(ntk4_file,"%s %s %s\n",ptokens[numtokens-3],
                ptokens[numtokens-2],ptokens[numtokens-1]);
              fflush(ntk4_file);

            } /* END WHILE more lines in ntk3 file */
            
          /* ELSE */
          } else {

            /* copy line as is into ntk4 file */
            fprintf(ntk4_file,"%s",network_spec_copy); fflush(ntk4_file);

          /* END IF */
          }

        /* ELSE (character is c or .) */
        } else {

          /* copy line as is into ntk4 file */
          fprintf(ntk4_file,"%s",network_spec); fflush(ntk4_file);

        } /* END IF */

      } /* END WHILE */

      num_leafs--;

      /* move ntk4 to ntk2 */
      fclose(ntk4_file);
      fclose(ntk3_file);
      fclose(ntk2_file);
      system("mv tempntk4 tempntk2\n");
      system("rm tempntk3\n");

    } /* END WHILE more leaf cells remain */

    if (debug>=10) {
      printf("\n    END OF .NTK FILE\n\n"); fflush(stdout);
    }

    if ((ntk2_file = fopen("tempntk2","r"))==NULL) {
      printf("\n  interpret_ntk_file:  unable to open file:  tempntk2\n");
      printf("  Exiting.\n");
      fflush(stdout); exit(1);
    }

    /* WHILE more lines exist */
    while (fgets(network_spec,MAX_CHARS_PER_LINE,ntk2_file) != NULL && !end_data_found) {

      if (debug>=10) {
        printf("    %s",network_spec); fflush(stdout);
      }

      /* read a line */
      strcpy(network_spec_copy,network_spec);

      /* get first token */
      ptokens[0] = (char *) strtok(network_spec," ,\n");

      /* IF first token is "h" */
      if (strcmp(ptokens[0],"h")==0) {

        /* parse remaining tokens */
        numtokens=1;
        while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n"))
                 !=NULL) numtokens++;

        /* copy line to appropriate buffer */
        if (strcmp(ptokens[1],"INPUT")==0) {
          if (input_num>=MAX_NUM_INPUTS) {
            printf("\n  Error:  attempt to define too many INPUTS");
                    printf(" (max = %d)\n", MAX_NUM_INPUTS);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          strcpy(buffer_inputs[input_num],ptokens[4]);
          strcat(buffer_inputs[input_num]," ");
          strcat(buffer_inputs[input_num],ptokens[3]);
          input_num++;
        }
        if (strcmp(ptokens[1],"NEURON")==0) {
          if (neuron_num>=MAX_NUM_NEURONS) {
            printf("\n  Error:  attempt to define too many NEURONS");
                    printf(" (max = %d)\n", MAX_NUM_NEURONS);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          strcpy(buffer_neurons[neuron_num],ptokens[5]);
          strcat(buffer_neurons[neuron_num]," ");
          strcat(buffer_neurons[neuron_num],ptokens[3]);
          strcat(buffer_neurons[neuron_num]," ");
          strcat(buffer_neurons[neuron_num],ptokens[4]);
          strcat(buffer_neurons[neuron_num]," ");
          strcat(buffer_neurons[neuron_num],ptokens[6]);
          neuron_num++;
        }
        if (strcmp(ptokens[1],"BUFFER")==0) {
          if (buffer_num>=MAX_NUM_BUFFERS) {
            printf("\n  Error:  attempt to define too many BUFFERS");
                    printf(" (max = %d)\n", MAX_NUM_BUFFERS);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          strcpy(buffer_buffers[buffer_num],ptokens[5]);
          strcat(buffer_buffers[buffer_num]," ");
          strcat(buffer_buffers[buffer_num],ptokens[3]);
          strcat(buffer_buffers[buffer_num]," ");
          strcat(buffer_buffers[buffer_num],ptokens[4]);
          if (numtokens == 10) {
            strcat(buffer_buffers[buffer_num]," ");
            strcat(buffer_buffers[buffer_num],ptokens[6]);
          }
          buffer_num++;
        }
        if (strcmp(ptokens[1],"SYNEXC")==0 ||
            strcmp(ptokens[1],"SYNINH")==0) {
          if (synapse_num>=MAX_NUM_SYNAPSES) {
            printf("\n  Error:  attempt to define too many SYNAPSES");
                    printf(" (max = %d)\n", MAX_NUM_SYNAPSES);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          strcpy(buffer_synapses[synapse_num],ptokens[3]);
          strcat(buffer_synapses[synapse_num]," ");
          strcat(buffer_synapses[synapse_num],ptokens[4]);
          strcat(buffer_synapses[synapse_num]," ");
          strcat(buffer_synapses[synapse_num],ptokens[5]);
          strcat(buffer_synapses[synapse_num]," ");
          strcat(buffer_synapses[synapse_num],ptokens[6]);
          if (numtokens == 11) {
            strcat(buffer_synapses[synapse_num]," ");
            strcat(buffer_synapses[synapse_num],ptokens[7]);
          }
          synapse_num++;
        }

        if (strcmp(ptokens[1],"TONICEXC")==0 ||
            strcmp(ptokens[1],"TONICINH")==0) {
          
          /* store parameters */
          if (tonic_num>=MAX_NUM_TONICS) {
            printf("\n  Error:  attempt to define too many ");
            printf("tonic inputs (max = %d)\n", MAX_NUM_TONICS);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          tonic[tonic_num].id = tonic_num;
          strcpy(tonic[tonic_num].to_node_string,ptokens[3]);
          strcpy(tonic[tonic_num].i_tonic_string,ptokens[4]);

          /* increment tonic_num */
          tonic_num++;

        }
        data_found = TRUE;
      }
      /* END IF first token is "h" */

      /* IF first token is "." */
      if (strcmp(ptokens[0],".")==0) {

        if (data_found) end_data_found = TRUE;

      }
      /* END IF first token is "." */

    } /* END WHILE more lines exist */

    if (debug>=5) {
      printf("\n  input strings collected:\n"); fflush(stdout);
      for (i=0;i<input_num;i++) {
        printf("  %s\n",buffer_inputs[i]); fflush(stdout);
      }
      printf("\n  neuron strings collected:\n"); fflush(stdout);
      for (i=0;i<neuron_num;i++) {
        printf("  %s\n",buffer_neurons[i]); fflush(stdout);
      }
      printf("\n  buffer strings collected:\n"); fflush(stdout);
      for (i=0;i<buffer_num;i++) {
        printf("  %s\n",buffer_buffers[i]); fflush(stdout);
      }
      printf("\n  synapse strings collected:\n"); fflush(stdout);
      for (i=0;i<synapse_num;i++) {
        printf("  %s\n",buffer_synapses[i]); fflush(stdout);
      }
    }

    num_inputs = input_num;
    num_neurons = neuron_num;
    num_buffers = buffer_num;
    num_synapses = synapse_num;
    input_num = 0;
    neuron_num = 0;
    buffer_num = 0;
    synapse_num = 0;

    if (debug>=10) {
      printf("\n  sorting input strings:\n"); 
      printf("  Number of input strings: %d\n",num_inputs); fflush(stdout);
    }

    /* pull out the inputs in sorted order */
    /* FOR all strings in the buffer */
    for (i=num_inputs-1;i>=0;i--) {

      if (debug>=10) {
        printf("\n  finding the best line\n"); fflush(stdout);
      }

      /* find the best line */
      best_index = 0;
      if (i>0) {
        for (j=1;j<=i;j++) {
          if (strcmp(buffer_inputs[best_index],buffer_inputs[j])>0) {
            best_index=j;
          }
        }
      }

      /* parse into tokens */
      ptokens[0] = (char *) strtok(buffer_inputs[best_index]," ,\n");
      numtokens=1;
      while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n")) 
         != NULL) numtokens++;

      if (debug>=10) {
        printf("  interpreting the line\n"); 
        printf("  best_index = %d\n",best_index); fflush(stdout);
        printf("  ptokens[0] = %s\n",ptokens[0]); 
        printf("  ptokens[1] = %s\n",ptokens[1]); 
        printf("  object_num = %d\n",object_num); fflush(stdout);
      }
      if (object_num>=MAX_NUM_OBJECTS) {
        printf("\n  Error:  attempt to define too many OBJECTS");
                printf(" (max = %d)\n", MAX_NUM_OBJECTS);
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }

      /* interpret the line */
      /* store to_node identifier for this object */
      strcpy(object_map[object_num].to_node_label,ptokens[1]);
      strcpy(object_map[object_num].from_node_label,"NOT_USED");

      /* store label for this object */
      if (!valid_label(ptokens[0])) {
        printf("  Error:  invalid label for INPUT: %s\n", ptokens[0]);
        fflush(stdout);
        num_errors_found++;
      }
      if (found_node_string(ptokens[0])) {
        printf("  Error:  INPUT label begins with \"node\"\n");
        printf("          This error may indicate an unconnected parameter FROM-gate\n");
        fflush(stdout);
        num_errors_found++;
      }
      strcpy(object_map[object_num].label,ptokens[0]);

      /* store parameters for this object */
      object_map[object_num].type = (int)INPUT;
      object_map[object_num].sub_id = input_num;

      /* increment object_num,input_num */
      object_num++;
      input_num++;

      /* clear the line by copying last entry into used position */
      strcpy(buffer_inputs[best_index],buffer_inputs[i]);

    } /* END FOR all strings in the buffer */

    if (debug>=10) {
      printf("\n  sorting neuron strings:\n"); 
      printf("  Number of neuron strings: %d\n",num_neurons); fflush(stdout);
    }

    /* pull out the neurons in sorted order */
    /* FOR all strings in the buffer */
    for (i=num_neurons-1;i>=0;i--) {

      if (debug>=10) {
        printf("\n  finding the best line\n"); fflush(stdout);
      }

      /* find the best line */
      best_index = 0;
      if (i>0) {
        for (j=1;j<=i;j++) {
          if (strcmp(buffer_neurons[best_index],buffer_neurons[j])>0) {
            best_index=j;
          }
        }
      }

      if (debug>=10) {
        printf("  parsing into tokens\n"); fflush(stdout);
      }

      /* parse into tokens */
      ptokens[0] = (char *) strtok(buffer_neurons[best_index]," ,\n");
      numtokens=1;
      while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n")) 
               != NULL) numtokens++;

      if (debug>=10) {
        printf("  interpreting the line\n"); 
        printf("  best_index = %d\n",best_index); fflush(stdout);
        printf("  ptokens[0] = %s\n",ptokens[0]); 
        printf("  ptokens[1] = %s\n",ptokens[1]); 
        printf("  ptokens[2] = %s\n",ptokens[2]);
        printf("  ptokens[3] = %s\n",ptokens[3]);
        printf("  object_num = %d\n",object_num); fflush(stdout);
      }

      if (object_num>=MAX_NUM_OBJECTS) {
        printf("\n  Error:  attempt to define too many OBJECTS");
                printf(" (max = %d)\n", MAX_NUM_OBJECTS);
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }

      /* interpret the line */
      /* store to_node and from_node identifiers for this object */
      strcpy(object_map[object_num].from_node_label,ptokens[1]);
      strcpy(object_map[object_num].to_node_label,ptokens[2]);

      if (debug>=10) {
        printf("  storing the label and parameters\n"); fflush(stdout);
      }

      /* store label for this object */
      if (!valid_label(ptokens[0])) {
        printf("  Error:  invalid label for NEURON: %s\n", ptokens[0]);
        fflush(stdout);
        num_errors_found++;
      }
      if (found_node_string(ptokens[0])) {
        printf("  Error:  NEURON label begins with \"node\"\n");
        printf("          This error may indicate an unconnected parameter FROM-gate\n");
        fflush(stdout);
        num_errors_found++;
      }
      strcpy(object_map[object_num].label,ptokens[0]);

      /* store parameters */
      if (neuron_num>=MAX_NUM_NEURONS) {
        printf("\n  interpret_ntk_file:  attempt to define too many neurons");
		printf(" (max = %d)\n", MAX_NUM_NEURONS);
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }
      neuron[neuron_num].id = object_num;
      neuron[neuron_num].t_refract = atof(ptokens[3]);
      if (!valid_number(ptokens[3])||neuron[neuron_num].t_refract<0) {
        printf("  Error:  invalid refractory period for NEURON: %s\n", 
          ptokens[3]);
        printf("          This error may indicate an unconnected parameter FROM-gate\n");
        fflush(stdout);
        num_errors_found++;
      }

      object_map[object_num].type = (int)NEURON;
      object_map[object_num].sub_id = neuron_num;

      if (debug>=10) {
        printf("  looking for tonic input\n"); fflush(stdout);
      }

      /* look for tonic inputs, add them up */
      neuron[neuron_num].i_tonic = 0.0;
      for (ii=0;ii<tonic_num;ii++) {
        if (strcmp(object_map[object_num].from_node_label,
                    tonic[ii].to_node_string)==0) { 
          if (debug>=10) {
            printf("  found tonic input, value = %s \n",
               tonic[ii].i_tonic_string); fflush(stdout);
          }
          neuron[neuron_num].i_tonic += atof(tonic[ii].i_tonic_string);
          if (!valid_number(&tonic[ii].i_tonic_string[0])) {
            printf("  Error:  invalid tonic input current for NEURON: %s\n", 
              tonic[ii].i_tonic_string);
            printf("          This error may indicate an unconnected parameter FROM-gate\n");
            fflush(stdout);
            num_errors_found++;
          }
          if (found_node_string(&tonic[ii].i_tonic_string[0])) {
            printf("  Error:  TONIC label begins with \"node\"\n");
            printf("          This error may indicate an unconnected parameter FROM-gate\n");
            fflush(stdout);
            num_errors_found++;
          }
        }
      }

      if (debug>=10) {
        printf("  total tonic input value = %f \n",
           neuron[neuron_num].i_tonic); fflush(stdout);
      }

      /* increment object_num, neuron_num */
      object_num++;
      neuron_num++;

      /* clear the line by copying last entry into used position */
      strcpy(buffer_neurons[best_index],buffer_neurons[i]);

    } /* END FOR all neuron lines */

    if (debug>=10) {
      printf("\n  sorting buffer strings:\n"); 
      printf("  Number of buffer strings: %d\n",num_buffers); fflush(stdout);
    }

    /* pull out the buffers in sorted order */
    /* FOR all strings in the buffer */
    for (i=num_buffers-1;i>=0;i--) {

      if (debug>=10) {
        printf("  finding the best line\n"); fflush(stdout);
      }

      /* find the best line */
      best_index = 0;
      if (i>0) {
        for (j=1;j<=i;j++) {
          if (strcmp(buffer_buffers[best_index],buffer_buffers[j])>0) {
            best_index=j;
          }
        }
      }

      if (debug>=10) {
        printf("  parsing into tokens\n"); fflush(stdout);
      }

      /* parse into tokens */
      ptokens[0] = (char *) strtok(buffer_buffers[best_index]," ,\n");
      numtokens=1;
      while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n")) 
         != NULL) numtokens++;

      if (debug>=10) {
        printf("  interpreting the line\n"); 
        printf("  best_index = %d\n",best_index); fflush(stdout);
        printf("  ptokens[0] = %s\n",ptokens[0]); 
        printf("  ptokens[1] = %s\n",ptokens[1]); 
        printf("  ptokens[2] = %s\n",ptokens[2]);
        if (numtokens == 4) {
          printf("  ptokens[3] = %s\n",ptokens[3]);
        }
        printf("  object_num = %d\n",object_num); fflush(stdout);
      }

      if (object_num>=MAX_NUM_OBJECTS) {
        printf("\n  Error:  attempt to define too many OBJECTS");
                printf(" (max = %d)\n", MAX_NUM_OBJECTS);
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }

      /* interpret the line */
      /* store from_node identifiers for this object */
      strcpy(object_map[object_num].from_node_label,ptokens[1]);
      strcpy(from_node_string,ptokens[1]);
      strcpy(object_map[object_num].to_node_label,ptokens[2]);

      if (debug>=10) {
        printf("  storing the label and parameters\n"); fflush(stdout);
      }

      /* store label for this object */
      if (!valid_label(ptokens[0])) {
        printf("  Error:  invalid label for BUFFER: %s\n", ptokens[0]);
        fflush(stdout);
        num_errors_found++;
      }
      if (found_node_string(ptokens[0])) {
        printf("  Error:  BUFFER label begins with \"node\"\n");
        printf("          This error may indicate an unconnected parameter FROM-gate\n");
        fflush(stdout);
        num_errors_found++;
      }
      strcpy(object_map[object_num].label,ptokens[0]);

      /* store parameters */
      if (buffer_num>=MAX_NUM_BUFFERS) {
        printf("\n  interpret_ntk_file:  attempt to define too many buffers");
		printf(" (max = %d)\n", MAX_NUM_BUFFERS);
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }
      buffer[buffer_num].id = object_num;
      if (numtokens == 4) {
        buffer[buffer_num].t_delay = atof(ptokens[3]);
        if (!valid_number(ptokens[3])||buffer[buffer_num].t_delay<0) {
          printf("  Error:  invalid delay time for BUFFER: %s\n", 
            ptokens[3]);
          printf("          This error may indicate an unconnected parameter FROM-gate\n");
          fflush(stdout);
          num_errors_found++;
        }
      } else {
        buffer[buffer_num].t_delay = 0.0;
      }

      if (debug>=10) {
        printf("  storing object type and sub_id\n"); fflush(stdout);
      }

      object_map[object_num].type = (int)BUFFER;
      object_map[object_num].sub_id = buffer_num;

      if (debug>=10) {
        printf("  searching inputs and neurons for matching node label\n");
        fflush(stdout);
      }

      /* search all inputs and neurons for matching node label */
      j=0;
      while (strcmp(from_node_string,object_map[j].to_node_label)!=0 &&
             j < input_num + neuron_num) j++;
      if (j < input_num + neuron_num) {
        from_node = j;
      } else {
        from_node = -1;
      }

      /* store parameters */
      buffer[buffer_num].id = object_num;
      buffer[buffer_num].object_from = from_node;

      if (debug>=10) {
        printf("%s\n",buffer_buffers[best_index]); fflush(stdout);
      }

      /* increment object_num, buffer_num */
      object_num++;
      buffer_num++;

      /* clear the line by copying last entry into used position */
      strcpy(buffer_buffers[best_index],buffer_buffers[i]);

    } /* END FOR all buffer lines */

    if (debug>=10) {
      printf("\n  sorting synapse strings:\n"); 
      printf("  Number of synapse strings: %d\n",num_synapses); fflush(stdout);
    }

    /* pull out the synapses */
    /* FOR all strings in the buffer */
    for (synapse_num=0;synapse_num<num_synapses;synapse_num++) {

      /* parse into tokens */
      ptokens[0] = (char *) strtok(buffer_synapses[synapse_num]," ,\n");
      numtokens=1;
      while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n")) 
              != NULL) numtokens++;

      /* interpret the line */

      if (object_num>=MAX_NUM_OBJECTS) {
        printf("\n  Error:  attempt to define too many OBJECTS");
                printf(" (max = %d)\n", MAX_NUM_OBJECTS);
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }

      /* parse out the to_node and from_node identifiers */
      strcpy(object_map[object_num].from_node_label,ptokens[0]);
      strcpy(from_node_string,ptokens[0]);
      strcpy(object_map[object_num].to_node_label,ptokens[1]);
      strcpy(to_node_string,ptokens[1]);

      /* find the corresponding object numbers */
      /* must search all inputs, neurons and buffers to find input for this */
	  /*   synapse */
      i=0;
      while (strcmp(from_node_string,object_map[i].to_node_label)!=0 &&
             i < input_num + neuron_num + buffer_num) i++;
      if (i < input_num + neuron_num + buffer_num) {
        from_node = i;
      } else {
        from_node = -1;   /* error condition, will catch it later */
      }

      /* only need to search neurons to find output for this synapse */
      i=input_num;
      while (strcmp(to_node_string,object_map[i].from_node_label)!=0 &&
             i < input_num + neuron_num) i++;
      if (i < input_num + neuron_num) {
        to_node = i;
      } else {
	    to_node = -1;
      }

      /* store parameters */
      if (synapse_num>=MAX_NUM_SYNAPSES) {
        printf("\n  interpret_ntk_file:  attempt to define too many synapses ");
        printf("(max = %d)\n",MAX_NUM_SYNAPSES);
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }
      synapse[synapse_num].id = object_num;
      synapse[synapse_num].object_from = from_node;
      synapse[synapse_num].object_to = to_node;
      synapse[synapse_num].i_weight = atof(ptokens[2]);
      if (!valid_number(ptokens[2])) {
        printf("  Error:  invalid weight current for SYNAPSE: %s\n", 
          ptokens[2]);
        fflush(stdout);
        num_errors_found++;
      }
      if (found_node_string(ptokens[2])) {
        printf("  Error:  SYNAPSE label begins with \"node\"\n");
        printf("          This error may indicate an unconnected parameter FROM-gate\n");
        fflush(stdout);
        num_errors_found++;
      }
      synapse[synapse_num].t_neuro = atof(ptokens[3]);
      if (!valid_number(ptokens[3])||synapse[synapse_num].t_neuro<0) {
        printf("  Error:  invalid post-synaptic current pulse time for SYNAPSE: %s\n", 
          ptokens[3]);
        fflush(stdout);
        num_errors_found++;
      }
      synapse[synapse_num].cliplevel = 1;  /* default value */
      if (numtokens==5) {
        synapse[synapse_num].cliplevel = atoi(ptokens[4]);
        if (!valid_number(ptokens[4])||synapse[synapse_num].cliplevel<1) {
          printf("  Error:  invalid clip level for SYNAPSE: %s\n", 
            ptokens[4]);
          fflush(stdout);
          num_errors_found++;
        }
      } 

      object_map[object_num].type = (int)SYNAPSE;
      object_map[object_num].sub_id = synapse_num;

      /* generate the object label for the synapse */
      strcpy(object_map[object_num].label,"s");
      strcat(object_map[object_num].label,object_map[from_node].label);
      strcat(object_map[object_num].label,object_map[to_node].label);

      /* increment object_num */
      object_num++;

    } /* END FOR all synapse lines */

    /* check for duplicate synapse labels */
    /* FOR all strings in the buffer */
    for (i=num_synapses-1;i>=0;i--) {
      duplicate_found = FALSE;
      suffix_string[0] = 'a';
      suffix_string[1] = '\0';
      for (j=0;j<i;j++) {
        if (strcmp(object_map[synapse[i].id].label,
                   object_map[synapse[j].id].label)==0) {
          duplicate_found = TRUE;
          strcat(object_map[synapse[j].id].label,suffix_string);
          suffix_string[0] += 1;
        }
      }
      if (duplicate_found) {
        strcat(object_map[synapse[i].id].label,suffix_string);
      }
    }

    /* close files */
    fclose(ntk_file);
    fclose(ntk2_file);
    if (debug <10)  system("rm tempntk2\n");

/* END */
}

/*****************************************************************/
void
check_for_network_errors()
/*****************************************************************/
{

    extern struct object_map_struct object_map[];
    extern struct neuron_struct neuron[];
    extern struct synapse_struct synapse[];
    extern struct tonic_struct tonic[];
    extern struct buffer_struct buffer[];
    extern int          neuron_num,synapse_num,input_num,object_num;
    extern int          debug, buffer_num;
    extern int          tonic_num;
    extern char        *object_type_string[];
    extern int          num_errors_found;

    int                 i,j;
    int                 num_neurons_found;
    int                 num_synapse_connections;


    if (debug>=5) {
      printf("  checking for errors in network specifications.\n"); 
      fflush(stdout);
    }

    /* check for errors in network specification */

    /* SYNAPSE inputs only come from NEURON, INPUT or BUFFER */
    /* SYNAPSE outputs only to NEURON */
    for (i=0; i<synapse_num; i++) {
      if (object_map[synapse[i].object_from].type != (int)NEURON &&
          object_map[synapse[i].object_from].type != (int)INPUT &&
          object_map[synapse[i].object_from].type != (int)BUFFER) {
        printf("  Error:  SYNAPSE %s receives input from %s %s.\n",
          object_map[synapse[i].id].label,
          object_type_string[object_map[synapse[i].object_from].type],
          object_map[synapse[i].object_from].label);
        fflush(stdout);
        num_errors_found++;
      }
      if (object_map[synapse[i].object_to].type != (int)NEURON) {
        printf("  Error:  SYNAPSE %s supplies output to %s %s.\n",
          object_map[synapse[i].id].label,
          object_type_string[object_map[synapse[i].object_to].type],
          object_map[synapse[i].object_to].label);
        fflush(stdout);
        num_errors_found++;
      }
    }

    /* SYNAPSE output must be connected to exactly one NEURON input */
    for (i=0; i<synapse_num; i++) {
      num_neurons_found = 0;
      for (j=0; j<neuron_num; j++) {
        if (strcmp(object_map[neuron[j].id].from_node_label,
                   object_map[synapse[i].id].to_node_label)==0) {
          num_neurons_found++;
        }
      }
      /* IF no neurons found */
      if (num_neurons_found==0) {

        /* floating synapse */
        printf("  Error:  SYNAPSE output not connected to any NEURON\n");
        fflush(stdout);
        num_errors_found++;

      /* ELSE */
      } else {

        /* IF more than one neuron found */
        if (num_neurons_found>1) {

          /* multiple neurons connected to synapse */
          printf("  Error:  SYNAPSE output connected to more than one NEURON\n");
          fflush(stdout);
          num_errors_found++;

        }
      }
    }

    /* TONIC symbol must be connected to exactly one NEURON input */
    for (i=0; i<tonic_num; i++) {
      num_neurons_found = 0;
      for (j=0; j<neuron_num; j++) {
        if (strcmp(object_map[neuron[j].id].from_node_label,
                   tonic[i].to_node_string)==0) {
          num_neurons_found++;
        }
      }
      /* IF no neurons found */
      if (num_neurons_found==0) {

        /* floating tonic */
        printf("  Error:  TONIC output not connected to any NEURON\n");
        fflush(stdout);
        num_errors_found++;

      /* ELSE */
      } else {

        /* IF more than one neuron found */
        if (num_neurons_found>1) {

          /* multiple neurons connected to synapse */
          printf("  Error:  TONIC output connected to more than one NEURON\n");
          fflush(stdout);
          num_errors_found++;

        }
      }
    }
    /* no duplicate object labels */
    for (i=0; i<object_num; i++) {
      for (j=i; j<object_num; j++) {
        if (strcmp(object_map[i].label,object_map[j].label)==0 && i!=j) {
          printf("  Error:  object label %s appears more than once.\n",
             object_map[i].label);
          fflush(stdout);
          num_errors_found++;
        }
      }
    }

    /* NEURON outputs not shared with any other output, or with tonic */
    for (i=0; i<neuron_num; i++) {
      for (j=0; j<object_num; j++) {
        if (strcmp(object_map[neuron[i].id].to_node_label,
                   object_map[j].to_node_label)==0 && neuron[i].id!=j) {
          printf("  Error:  NEURON %s shares output with %s %s.\n",
            object_map[neuron[i].id].label,
            object_type_string[object_map[j].type],
            object_map[j].label);
          fflush(stdout);
          num_errors_found++;
        }
      }
      for (j=0; j<tonic_num; j++) {
        if (strcmp(object_map[neuron[i].id].to_node_label,
                   tonic[j].to_node_string)==0) {
          printf("  Error:  NEURON %s output connected to tonic input.\n",
            object_map[neuron[i].id].label);
          fflush(stdout);
          num_errors_found++;
        }
      }
    }

    /* INPUT outputs not shared with any other output, or with tonic */
    for (i=0; i<input_num; i++) {
      for (j=0; j<object_num; j++) {
        if (strcmp(object_map[i].to_node_label,
                   object_map[j].to_node_label)==0 && i!=j) {
          printf("  Error:  INPUT %s shares output with %s %s.\n",
            object_map[i].label,
            object_type_string[object_map[j].type],
            object_map[j].label);
          fflush(stdout);
          num_errors_found++;
        }
      }
      for (j=0; j<tonic_num; j++) {
        if (strcmp(object_map[i].to_node_label,
                   tonic[j].to_node_string)==0) {
          printf("  Error:  INPUT %s symbol connected to tonic input.\n",
            object_map[i].label);
          fflush(stdout);
          num_errors_found++;
        }
      }
    }

    /* BUFFER outputs not shared with tonic */
    for (i=0; i<buffer_num; i++) {
      for (j=0; j<tonic_num; j++) {
        if (strcmp(object_map[buffer[i].id].to_node_label,
                   tonic[j].to_node_string)==0) {
          printf("  Error:  BUFFER %s symbol connected to tonic input.\n",
            object_map[buffer[i].id].label);
          fflush(stdout);
          num_errors_found++;
        }
      }
      for (j=0; j<synapse_num; j++) {
        if (strcmp(object_map[buffer[i].id].to_node_label,
                   object_map[synapse[j].id].to_node_label)==0) {
          printf("  Error:  BUFFER %s symbol connected to SYNAPSE output.\n",
            object_map[buffer[i].id].label);
          fflush(stdout);
          num_errors_found++;
        }
      }
    }

    /* NEURON inputs not shared with any spiking output */
    for (i=0; i<neuron_num; i++) {
      for (j=0; j<input_num+neuron_num+buffer_num; j++) {
        if (strcmp(object_map[neuron[i].id].from_node_label,
                   object_map[j].to_node_label)==0) {
          printf("  Error:  NEURON %s receives direct spiking input (no synapse).\n",
            object_map[neuron[i].id].label);
          fflush(stdout);
          num_errors_found++;
        }
      }
    }

    /* only one output connected to SYNAPSE input, and it must be a spiker */
    for (i=0; i<synapse_num; i++) {
      num_synapse_connections = 0;
      for (j=0; j<object_num; j++) {
        if (strcmp(object_map[synapse[i].id].from_node_label,
                   object_map[j].to_node_label)==0) {
          if (object_map[j].type==(int)SYNAPSE) {
            printf("  Error:  SYNAPSE %s receives input from SYNAPSE.\n",
              object_map[synapse[i].id].label);
            fflush(stdout);
            num_errors_found++;
          } else {
            num_synapse_connections++;
          }
        }
      }
      if (num_synapse_connections == 0) {
        printf("  Error:  SYNAPSE %s receives no spiking inputs.\n",
          object_map[synapse[i].id].label);
        fflush(stdout);
        num_errors_found++;
      }
      if (num_synapse_connections > 1) {
        printf("  Error:  SYNAPSE %s receives more than one spiking input.\n",
          object_map[synapse[i].id].label);
        fflush(stdout);
        num_errors_found++;
      }
    }
    if (num_errors_found==1) {
      printf("  1 error found.\n  Exiting.\n\n",num_errors_found);
      fflush(stdout);
      exit(1);
    }
    if (num_errors_found >1) {
      printf("  %d errors found.\n  Exiting.\n\n",num_errors_found);
      fflush(stdout);
      exit(1);
    }

/* END */
}

/*****************************************************************/
void
interpret_spk_file()
/*****************************************************************/
{

    extern struct object_map_struct object_map[];
    extern struct neuron_struct neuron[];
    extern struct synapse_struct synapse[];
    extern struct tonic_struct tonic[];
    extern struct buffer_struct buffer[];
    extern int          neuron_num,synapse_num,input_num,object_num;
    extern int          tonic_num;
    extern int          debug, buffer_num;
    extern char        *object_type_string[];
    extern FILE        *spk_file;
    extern char         spk_file_name[];
    extern int          spk_file_found;
    extern int          num_errors_found;


    FILE               *fopen();
    FILE               *spk2_file;
    FILE               *spk3_file;
    FILE               *spk4_file;
    char                network_spec[MAX_CHARS_PER_LINE];
    char                network_spec_copy[MAX_CHARS_PER_LINE];
    char                buffer_inputs[MAX_NUM_INPUTS][MAX_CHARS_PER_LINE];
    char                buffer_neurons[MAX_NUM_NEURONS][MAX_CHARS_PER_LINE];
    char                buffer_synapses[MAX_NUM_SYNAPSES][MAX_CHARS_PER_LINE];
    char                buffer_buffers[MAX_NUM_BUFFERS][MAX_CHARS_PER_LINE];
    char                from_node_string[MAX_CHARS_PER_LINE];
    char                to_node_string[MAX_CHARS_PER_LINE];
    char                stored_symbol_names[MAX_NUM_SYMBOL_TYPES][MAX_CHARS_PER_LINE];
    char               *ptokens[MAX_TOKENS_PER_LINE];
    char                suffix_string[SUFFIX_STRING_DIGITS1];
    char                instance_name[MAX_CHARS_PER_LINE];
    char                cell_type[MAX_CHARS_PER_LINE];
    char                expand_cell_parameters[MAX_NUM_EXPAND_CELL_PARAMETERS][MAX_CHARS_PER_LINE];
    char                instance_parameters[MAX_NUM_EXPAND_CELL_PARAMETERS][MAX_CHARS_PER_LINE];
    int                 ii,i,j,numtokens;
    int                 data_found,end_data_found;
    int                 num_inputs;
    int                 num_neurons;
    int                 num_synapses;
    int                 num_buffers;
    int                 from_node;
    int                 to_node;
    int                 best_index;
    int                 spk_read_state;
    int                 num_leafs;
    int                 num_toplevel;
    int                 num_parameters;
    int                 found_index;
    int                 duplicate_found;
    int                 match_found;
    int                 eof_found;
    int                 tokenval;
    int                 enddef_found;
    int                 num_symbol_types, instance_number;
    int                 num_instances[MAX_NUM_SYMBOL_TYPES];

    /* initialize basic values */
    data_found = FALSE;
    end_data_found = FALSE;

    /* open spk file */
    if ((spk_file = fopen(spk_file_name,"r"))==NULL) return;

    spk_file_found = TRUE;

    printf("  interpreting %s\n",spk_file_name); fflush(stdout);

    /* open spk2 file */
    if ((spk2_file = fopen("tempspk2","w+"))==NULL) {
      printf("\n  interpret_spk_file:  unable to open file:  tempspk2\n");
      printf("  Exiting.\n");
      fflush(stdout); exit(1);
    }

    num_leafs = 0;
    num_toplevel = 0;
    spk_read_state = (int)SPK_LOOKING;

    /* WHILE more lines exist */
    while (fgets(network_spec,MAX_CHARS_PER_LINE,spk_file) != NULL) {

      if (debug>=10) {
        printf("    %s",network_spec); fflush(stdout);
      }

      /* copy spk line over to spk2 file if it is not blank and first char */
      /*        is not '#'                                                 */

      /* read a line */
      strcpy(network_spec_copy,network_spec);

      /* parse into tokens */
      ptokens[0] = (char *) strtok(network_spec," ,\n");

      /* IF line is not blank */
      if (ptokens[0]!=NULL) {

        /* IF line is not a comment */
        if (*ptokens[0]!='#') {
     
          /* finish parsing */
          numtokens=1;
          while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n"))
                   != NULL) numtokens++;
          
          tokenval = (int)SPK_OTHER;
          if (strcmp(ptokens[0],"def")==0) tokenval = (int)SPK_DEF;
          if (strcmp(ptokens[0],"begin")==0) tokenval = (int)SPK_BEGIN;
          if (strcmp(ptokens[0],"end")==0) tokenval = (int)SPK_END;
          if (strcmp(ptokens[0],"enddef")==0) tokenval = (int)SPK_ENDDEF;

          switch (tokenval) {

            case SPK_DEF:

              switch (spk_read_state) {
                case SPK_LOOKING:
                  spk_read_state = (int)SPK_GOT_DEF;
                  num_leafs++;
                  num_symbol_types = 0;
                  fprintf(spk2_file,"%s",network_spec_copy); 
                  fflush(spk2_file);
                  break;
                case SPK_GOT_BEGIN:
                case SPK_GOT_DEF:
                case SPK_GOT_TOPLEVEL:
                  printf("\n  unexpected def command in .spk file\n");
                  printf("  Exiting.\n");
                  fflush(stdout); exit(1);
                  break;
              }
              break;

            case SPK_ENDDEF:

              switch (spk_read_state) {
                case SPK_GOT_DEF:
                  spk_read_state = (int)SPK_LOOKING;
                  fprintf(spk2_file,"%s",network_spec_copy); 
                  fflush(spk2_file);
                  break;
                case SPK_GOT_BEGIN:
                case SPK_LOOKING:
                case SPK_GOT_TOPLEVEL:
                  printf("\n  unexpected enddef command in .spk file\n");
                  printf("  Exiting.\n");
                  fflush(stdout); exit(1);
                  break;
              }
              break;

            case SPK_BEGIN:

              switch (spk_read_state) {
                case SPK_LOOKING:
                  if (num_toplevel==0) {
                    num_toplevel++;
                    spk_read_state = (int)SPK_GOT_BEGIN;
                    num_symbol_types = 0;
                    fprintf(spk2_file,"%s",network_spec_copy); 
                    fflush(spk2_file);
                  } else {
                    printf("\n  unexpected enddef command in .spk file\n");
                    printf("  Exiting.\n");
                    fflush(stdout); exit(1);
                  }
                  break;
                case SPK_GOT_BEGIN:
                case SPK_GOT_DEF:
                case SPK_GOT_TOPLEVEL:
                  printf("\n  unexpected begin command in .spk file\n");
                  printf("  Exiting.\n");
                  fflush(stdout); exit(1);
                  break;
              }
              break;

            case SPK_END:

              switch (spk_read_state) {
                case SPK_GOT_BEGIN:
                  spk_read_state = (int)SPK_GOT_TOPLEVEL;
                  fprintf(spk2_file,"%s",network_spec_copy); 
                  fflush(spk2_file);
                  break;
                case SPK_LOOKING:
                case SPK_GOT_DEF:
                case SPK_GOT_TOPLEVEL:
                  printf("\n  unexpected end command in .spk file\n");
                  printf("  Exiting.\n");
                  fflush(stdout); exit(1);
                  break;
              }
              break;

            case SPK_OTHER:

              switch (spk_read_state) {
                case SPK_GOT_DEF:
                case SPK_GOT_BEGIN:
                 
                  /* compare first token with existing list */
                  match_found = FALSE;
                  if (num_symbol_types > 0) {
                    i=0;
                    while (strcmp(ptokens[0],stored_symbol_names[i])!=0 &&
                           i<num_symbol_types) i++;
                    if (i!=num_symbol_types) match_found = TRUE;
                  }
                  
                  /* IF it's already in the list, increment counter */
                  if (match_found) {
                    num_instances[i] += 1;
                    instance_number = num_instances[i];
                  } else {
                    /* else add it to list, set counter to one */
                    num_instances[num_symbol_types]=1;
                    strcpy(stored_symbol_names[num_symbol_types],ptokens[0]);
                    instance_number = 1;
                    num_symbol_types++;
                  }

                  /* concatenate token string and number to make 2nd token */
                  /* concatenate the remaining tokens */
                  fprintf(spk2_file,"%s %s%d ",ptokens[0],ptokens[0],
                    instance_number); 
                  for (i=1;i<numtokens;i++) {
                    fprintf(spk2_file,"%s ",ptokens[i]);
                  }
                  fprintf(spk2_file,"\n");
                  fflush(spk2_file);
                  break;
                case SPK_LOOKING:
                case SPK_GOT_TOPLEVEL:
                  printf("\n  unexpected end command in .spk file\n");
                  printf("  Exiting.\n");
                  fflush(stdout); exit(1);
                  break;
              }
              break;

          } /* SWITCH tokenval */
        } /* END IF line is not a comment */
      } /* END IF line is not blank */
    } /* END WHILE more lines exist */

    if (debug>=1 && num_leafs>0) {
      printf("  flattening hierarchy\n"); 
      fflush(stdout);
    }

    if (debug>=5) {
      printf("\n  number of leaf cells found = %d.\n\n",num_leafs); 
      fflush(stdout);
    }

    /* go back to beginning of spk2 file */
    fclose(spk2_file);

    /* next pass throught spk2 file */
    /* WHILE more leaf cells remain */
    while (num_leafs>0) {

      if ((spk2_file = fopen("tempspk2","r"))==NULL) {
        printf("\n  interpret_spk_file:  unable to open file:  tempspk2\n");
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }

      if ((spk3_file = fopen("tempspk3","w+"))==NULL) {
        printf("\n  interpret_spk_file:  unable to open file:  tempspk3\n");
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }

      if ((spk4_file = fopen("tempspk4","w+"))==NULL) {
        printf("\n  interpret_spk_file:  unable to open file:  tempspk4\n");
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }

      /* read first line of spk2 file */
      fgets(network_spec,MAX_CHARS_PER_LINE,spk2_file); 
      
      /* parse the line  */
      ptokens[0] = (char *) strtok(network_spec," ,\n");
  
      numtokens=1;
      while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n")) 
               != NULL) numtokens++;
  
      strcpy(cell_type,ptokens[1]);
      if (debug>=10) {
        printf("   expanding cell type:  %s\n",cell_type); 
        fflush(stdout);
      }

      num_parameters = 0;
      for (i=2; i<numtokens; i++) {
        strcpy(expand_cell_parameters[num_parameters],ptokens[i]);
        if (debug>=10) {
          printf("          parameters found:  %s\n",
             expand_cell_parameters[num_parameters]); 
          fflush(stdout);
        }
        num_parameters++;
      }

      /* copy following lines to spk3 until enddef */
      /* WHILE more lines exist AND first token <> "enddef" */
      eof_found = FALSE;
      enddef_found = FALSE;
      while (!eof_found && !enddef_found) {
 
        if (fgets(network_spec,MAX_CHARS_PER_LINE,spk2_file) == NULL) {
          eof_found = TRUE;
        } else {
          if (debug>=10) {
            printf("    %s",network_spec); fflush(stdout);
          }
          strcpy(network_spec_copy, network_spec);

          ptokens[0] = (char *) strtok(network_spec," ,\n");
          if (strcmp(ptokens[0],"enddef")!=0) {

            /* copy spk line over to spk3 file */
            fprintf(spk3_file,"%s",network_spec_copy); fflush(spk3_file);
          } else {
            enddef_found = TRUE;
          }
        }
      } /* END WHILE more lines exist */

      /* expand leaf cell */

      /* WHILE more lines exist */
      while (fgets(network_spec,MAX_CHARS_PER_LINE,spk2_file) != NULL) {

        if (debug>=10) {
          printf("    %s",network_spec); fflush(stdout);
        }

        /* parse the line */
        strcpy(network_spec_copy,network_spec);
        ptokens[0] = (char *) strtok(network_spec," ,\n");
 
        numtokens=1;
        while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n")) 
                 != NULL) numtokens++;

        /* IF this line is an instance of the cell to be expanded */
        if (strcmp(ptokens[0],cell_type)==0) {

          /* expand it */
          /* store instance name and parameters */
          strcpy(instance_name,ptokens[1]);
          for (i=0;i<num_parameters;i++) {
            strcpy(instance_parameters[i],ptokens[i+2]);
          }
     
          /* go back to beginning of spk3 file */
          fseek(spk3_file, (long) 0, 0);

          /* WHILE more lines in spk3 file */
          while (fgets(network_spec,MAX_CHARS_PER_LINE,spk3_file) 
                  != NULL) {

            /* parse the line */
            ptokens[0] = (char *) strtok(network_spec," ,\n");
            numtokens=1;
            while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n"))
                     != NULL) numtokens++;

            /* write out first three words */
            fprintf(spk4_file,"%s %s%s ",ptokens[0],
              instance_name,ptokens[1]); fflush(spk4_file);
         
            /* FOR each parameter */
            for (i=2; i<numtokens; i++) {

              /* IF this parameter is an object label (i.e. n1, b1, or in1) */
              if ((strcmp(ptokens[0],"NEURON")==0 && i==4) ||
                  (strcmp(ptokens[0],"BUFFER")==0 && i==4) ||
                  (strcmp(ptokens[0],"INPUT")==0 && i==3)) {

                /* prepend instance_name and write it */
                fprintf(spk4_file,"%s%s ",instance_name,ptokens[i]);
                fflush(spk4_file);
                  
              /* ELSE */
              } else {

                /* IF numerical value (first char is digit . + -) */
                if (*ptokens[i]=='.' || isdigit((int)*ptokens[i]) || 
                    *ptokens[i]=='-' || *ptokens[i]=='+') {

                  /* write it as is */
                  fprintf(spk4_file,"%s ",ptokens[i]);
                  fflush(spk4_file);

                /* ELSE */
                } else {

                  /* look for this parameter in the instance parameter list */
                  found_index=-1;
                  for (j=0;j<num_parameters;j++) {
                    if (strcmp(ptokens[i],expand_cell_parameters[j]) == 0) 
                      found_index = j;
                  }
 
                  /* IF parameter found */
                  if (found_index != -1) {

                    /* write out replacement parameter */
                    fprintf(spk4_file,"%s ",instance_parameters[found_index]);
                    fflush(spk4_file);

                  /* ELSE */
                  } else {

                    /* prepend instance_name and write it */
                    fprintf(spk4_file,"%s%s ",instance_name,ptokens[i]);
                    fflush(spk4_file);

                  /* END IF parameter found */
                  }
                /* END IF numerical value */
                }
              /* END IF parameter is object label */
              }
            /* END FOR each parameter */
            } 

            /* write out \n */
            fprintf(spk4_file,"\n");
            fflush(spk4_file);

          } /* END WHILE more lines in spk3 file */
            
        /* ELSE */
        } else {

          /* copy line as is into spk4 file */
          fprintf(spk4_file,"%s",network_spec_copy); fflush(spk4_file);

        /* END IF */
        }

      } /* END WHILE */

      num_leafs--;

      /* move spk4 to spk2 */
      fclose(spk4_file);
      fclose(spk3_file);
      fclose(spk2_file);
      system("mv tempspk4 tempspk2\n");
      system("rm tempspk3\n");

    } /* END WHILE more leaf cells remain */

    if (debug>=10) {
      printf("\n    END OF .NTK FILE\n\n"); fflush(stdout);
    }

    if ((spk2_file = fopen("tempspk2","r"))==NULL) {
      printf("\n  interpret_spk_file:  unable to open file:  tempspk2\n");
      printf("  Exiting.\n");
      fflush(stdout); exit(1);
    }

    /* WHILE more lines exist */
    while (fgets(network_spec,MAX_CHARS_PER_LINE,spk2_file) != NULL && !end_data_found) {

      if (debug>=10) {
        printf("    %s",network_spec); fflush(stdout);
      }

      /* read a line */
      strcpy(network_spec_copy,network_spec);

      /* get first token */
      ptokens[0] = (char *) strtok(network_spec," ,\n");

      /* IF first token is not "begin" or "end" */
      if (strcmp(ptokens[0],"begin")!=0 && strcmp(ptokens[0],"end")!=0) {

        /* parse remaining tokens */
        numtokens=1;
        while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n"))
                 !=NULL) numtokens++;

        /* copy line to appropriate buffer */
        if (strcmp(ptokens[0],"INPUT")==0) {
          if (input_num>=MAX_NUM_INPUTS) {
            printf("\n  Error:  attempt to define too many INPUTS");
		    printf(" (max = %d)\n", MAX_NUM_INPUTS);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          if (numtokens>=4) {
            strcpy(buffer_inputs[input_num],ptokens[3]);
            strcat(buffer_inputs[input_num]," ");
            strcat(buffer_inputs[input_num],ptokens[2]);
          } else {
            printf("\n  Error:  not enough parameters for INPUT in .spk file");
            printf("\n  Exiting.\n");
            fflush(stdout); exit(1);
          }
          input_num++;
        }
        if (strcmp(ptokens[0],"NEURON")==0) {
          if (neuron_num>=MAX_NUM_NEURONS) {
            printf("\n  Error:  attempt to define too many NEURONS");
		    printf(" (max = %d)\n", MAX_NUM_NEURONS);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          if (numtokens>=6) {
            strcpy(buffer_neurons[neuron_num],ptokens[4]);
            strcat(buffer_neurons[neuron_num]," ");
            strcat(buffer_neurons[neuron_num],ptokens[2]);
            strcat(buffer_neurons[neuron_num]," ");
            strcat(buffer_neurons[neuron_num],ptokens[3]);
            strcat(buffer_neurons[neuron_num]," ");
            strcat(buffer_neurons[neuron_num],ptokens[5]);
          } else {
            printf("\n  Error:  not enough parameters for NEURON in .spk file");
            printf("\n  Exiting.\n");
            fflush(stdout); exit(1);
          }
          neuron_num++;
        }
        if (strcmp(ptokens[0],"BUFFER")==0) {
          if (buffer_num>=MAX_NUM_BUFFERS) {
            printf("\n  Error:  attempt to define too many BUFFERS");
		    printf(" (max = %d)\n", MAX_NUM_BUFFERS);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          if (numtokens>=5) {
            strcpy(buffer_buffers[buffer_num],ptokens[4]);
            strcat(buffer_buffers[buffer_num]," ");
            strcat(buffer_buffers[buffer_num],ptokens[2]);
            strcat(buffer_buffers[buffer_num]," ");
            strcat(buffer_buffers[buffer_num],ptokens[3]);
            if (numtokens == 6) {
              strcat(buffer_buffers[buffer_num]," ");
              strcat(buffer_buffers[buffer_num],ptokens[5]);
            }
          } else {
            printf("\n  Error:  not enough parameters for BUFFER in .spk file");
            printf("\n  Exiting.\n");
            fflush(stdout); exit(1);
          }
          buffer_num++;
        }
        if (strcmp(ptokens[0],"SYNEXC")==0 ||
            strcmp(ptokens[0],"SYNINH")==0 ||
            strcmp(ptokens[0],"SYNAPSE")==0) {
          if (synapse_num>=MAX_NUM_SYNAPSES) {
            printf("\n  Error:  attempt to define too many SYNAPSES");
		    printf(" (max = %d)\n", MAX_NUM_SYNAPSES);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          if (numtokens>=6) {
            strcpy(buffer_synapses[synapse_num],ptokens[2]);
            strcat(buffer_synapses[synapse_num]," ");
            strcat(buffer_synapses[synapse_num],ptokens[3]);
            strcat(buffer_synapses[synapse_num]," ");
            strcat(buffer_synapses[synapse_num],ptokens[4]);
            strcat(buffer_synapses[synapse_num]," ");
            strcat(buffer_synapses[synapse_num],ptokens[5]);
            if (numtokens == 7) {
              strcat(buffer_synapses[synapse_num]," ");
              strcat(buffer_synapses[synapse_num],ptokens[6]);
            }
          } else {
            printf("\n  Error:  not enough parameters for SYNAPSE in .spk file");
            printf("\n  Exiting.\n");
            fflush(stdout); exit(1);
          }
          synapse_num++;
        }

        if (strcmp(ptokens[0],"TONICEXC")==0 ||
            strcmp(ptokens[0],"TONICINH")==0 ||
            strcmp(ptokens[0],"TONIC")==0) {
          
          /* store parameters */
          if (tonic_num>=MAX_NUM_TONICS) {
            printf("\n  Error:  attempt to define too many ");
            printf("tonic inputs (max = %d)\n", MAX_NUM_TONICS);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          if (numtokens>=4) {
            tonic[tonic_num].id = tonic_num;
            strcpy(tonic[tonic_num].to_node_string,ptokens[2]);
            strcpy(tonic[tonic_num].i_tonic_string,ptokens[3]);
          } else {
            printf("\n  Error:  not enough parameters for TONIC in .spk file");
            printf("\n  Exiting.\n");
            fflush(stdout); exit(1);

          }

          /* increment tonic_num */
          tonic_num++;

        }
        data_found = TRUE;

      /* END IF first token is not "begin" or "end" */
      } else {

        /* IF first token is "end" */
        if (strcmp(ptokens[0],"end")==0) {

          if (data_found) end_data_found = TRUE;

        /* END IF first token is "end" */
        }
      }
    } /* END WHILE more lines exist */

    if (debug>=5) {
      printf("\n  input strings collected:\n"); fflush(stdout);
      for (i=0;i<input_num;i++) {
        printf("  %s\n",buffer_inputs[i]); fflush(stdout);
      }
      printf("\n  neuron strings collected:\n"); fflush(stdout);
      for (i=0;i<neuron_num;i++) {
        printf("  %s\n",buffer_neurons[i]); fflush(stdout);
      }
      printf("\n  buffer strings collected:\n"); fflush(stdout);
      for (i=0;i<buffer_num;i++) {
        printf("  %s\n",buffer_buffers[i]); fflush(stdout);
      }
      printf("\n  synapse strings collected:\n"); fflush(stdout);
      for (i=0;i<synapse_num;i++) {
        printf("  %s\n",buffer_synapses[i]); fflush(stdout);
      }
    }

    num_inputs = input_num;
    num_neurons = neuron_num;
    num_buffers = buffer_num;
    num_synapses = synapse_num;
    input_num = 0;
    neuron_num = 0;
    buffer_num = 0;
    synapse_num = 0;

    if (debug>=10) {
      printf("\n  sorting input strings:\n"); 
      printf("  Number of input strings: %d\n",num_inputs); fflush(stdout);
    }

    /* pull out the inputs in sorted order */
    /* FOR all strings in the buffer */
    for (i=num_inputs-1;i>=0;i--) {

      if (debug>=10) {
        printf("\n  finding the best line\n"); fflush(stdout);
      }

      /* find the best line */
      best_index = 0;
      if (i>0) {
        for (j=1;j<=i;j++) {
          if (strcmp(buffer_inputs[best_index],buffer_inputs[j])>0) {
            best_index=j;
          }
        }
      }

      /* parse into tokens */
      ptokens[0] = (char *) strtok(buffer_inputs[best_index]," ,\n");
      numtokens=1;
      while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n")) 
         != NULL) numtokens++;

      if (debug>=10) {
        printf("  interpreting the line\n"); 
        printf("  best_index = %d\n",best_index); fflush(stdout);
        printf("  ptokens[0] = %s\n",ptokens[0]); 
        printf("  ptokens[1] = %s\n",ptokens[1]); 
        printf("  object_num = %d\n",object_num); fflush(stdout);
      }

      if (object_num>=MAX_NUM_OBJECTS) {
        printf("\n  Error:  attempt to define too many OBJECTS");
                printf(" (max = %d)\n", MAX_NUM_OBJECTS);
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }
      /* interpret the line */
      /* store to_node identifier for this object */
      strcpy(object_map[object_num].to_node_label,ptokens[1]);
      strcpy(object_map[object_num].from_node_label,"NOT_USED");

      /* store label for this object */
      if (!valid_label(ptokens[0])) {
        printf("  Error:  invalid label for INPUT: %s\n", ptokens[0]);
        fflush(stdout);
        num_errors_found++;
      }
      strcpy(object_map[object_num].label,ptokens[0]);

      /* store parameters for this object */
      object_map[object_num].type = (int)INPUT;
      object_map[object_num].sub_id = input_num;

      /* increment object_num,input_num */
      object_num++;
      input_num++;

      /* clear the line by copying last entry into used position */
      strcpy(buffer_inputs[best_index],buffer_inputs[i]);

    } /* END FOR all strings in the buffer */

    if (debug>=10) {
      printf("\n  sorting neuron strings:\n"); 
      printf("  Number of neuron strings: %d\n",num_neurons); fflush(stdout);
    }

    /* pull out the neurons in sorted order */
    /* FOR all strings in the buffer */
    for (i=num_neurons-1;i>=0;i--) {

      if (debug>=10) {
        printf("\n  finding the best line\n"); fflush(stdout);
      }

      /* find the best line */
      best_index = 0;
      if (i>0) {
        for (j=1;j<=i;j++) {
          if (strcmp(buffer_neurons[best_index],buffer_neurons[j])>0) {
            best_index=j;
          }
        }
      }

      if (debug>=10) {
        printf("  parsing into tokens\n"); fflush(stdout);
      }

      /* parse into tokens */
      ptokens[0] = (char *) strtok(buffer_neurons[best_index]," ,\n");
      numtokens=1;
      while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n")) 
               != NULL) numtokens++;

      if (debug>=10) {
        printf("  interpreting the line\n"); 
        printf("  best_index = %d\n",best_index); fflush(stdout);
        printf("  ptokens[0] = %s\n",ptokens[0]); 
        printf("  ptokens[1] = %s\n",ptokens[1]); 
        printf("  ptokens[2] = %s\n",ptokens[2]);
        printf("  ptokens[3] = %s\n",ptokens[3]);
        printf("  object_num = %d\n",object_num); fflush(stdout);
      }

      if (object_num>=MAX_NUM_OBJECTS) {
        printf("\n  Error:  attempt to define too many OBJECTS");
                printf(" (max = %d)\n", MAX_NUM_OBJECTS);
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }

      /* interpret the line */
      /* store to_node and from_node identifiers for this object */
      strcpy(object_map[object_num].from_node_label,ptokens[1]);
      strcpy(object_map[object_num].to_node_label,ptokens[2]);

      if (debug>=10) {
        printf("  storing the label and parameters\n"); fflush(stdout);
      }

      /* store label for this object */
      if (!valid_label(ptokens[0])) {
        printf("  Error:  invalid label for NEURON: %s\n", ptokens[0]);
        fflush(stdout);
        num_errors_found++;
      }
      strcpy(object_map[object_num].label,ptokens[0]);

      /* store parameters */
      if (neuron_num>=MAX_NUM_NEURONS) {
        printf("\n  interpret_spk_file:  attempt to define too many neurons");
		printf(" (max = %d)\n", MAX_NUM_NEURONS);
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }
      neuron[neuron_num].id = object_num;
      neuron[neuron_num].t_refract = atof(ptokens[3]);
      if (!valid_number(ptokens[3])||neuron[neuron_num].t_refract<0) {
        printf("  Error:  invalid refractory period for NEURON: %s\n", 
          ptokens[3]);
        fflush(stdout);
        num_errors_found++;
      }

      object_map[object_num].type = (int)NEURON;
      object_map[object_num].sub_id = neuron_num;

      if (debug>=10) {
        printf("  looking for tonic input\n"); fflush(stdout);
      }

      /* look for tonic input */
      neuron[neuron_num].i_tonic = 0.0;
      for (ii=0;ii<tonic_num;ii++) {
        if (strcmp(object_map[object_num].from_node_label,
                    tonic[ii].to_node_string)==0) { 
          if (debug>=10) {
            printf("  found tonic input, value = %s \n",
               tonic[ii].i_tonic_string); fflush(stdout);
          }
          neuron[neuron_num].i_tonic += atof(tonic[ii].i_tonic_string);
          if (!valid_number(&tonic[ii].i_tonic_string[0])) {
            printf("  Error:  invalid tonic input current for NEURON: %s\n", 
              tonic[ii].i_tonic_string);
            fflush(stdout);
            num_errors_found++;
          }
        }
      }

      if (debug>=10) {
        printf("  total tonic input value = %f \n",
           neuron[neuron_num].i_tonic); fflush(stdout);
      }

      /* increment object_num, neuron_num */
      object_num++;
      neuron_num++;

      /* clear the line by copying last entry into used position */
      strcpy(buffer_neurons[best_index],buffer_neurons[i]);

    } /* END FOR all neuron lines */

    if (debug>=10) {
      printf("\n  sorting buffer strings:\n"); 
      printf("  Number of buffer strings: %d\n",num_buffers); fflush(stdout);
    }

    /* pull out the buffers in sorted order */
    /* FOR all strings in the buffer */
    for (i=num_buffers-1;i>=0;i--) {

      if (debug>=10) {
        printf("  finding the best line\n"); fflush(stdout);
      }

      /* find the best line */
      best_index = 0;
      if (i>0) {
        for (j=1;j<=i;j++) {
          if (strcmp(buffer_buffers[best_index],buffer_buffers[j])>0) {
            best_index=j;
          }
        }
      }

      if (debug>=10) {
        printf("  parsing into tokens\n"); fflush(stdout);
      }

      /* parse into tokens */
      ptokens[0] = (char *) strtok(buffer_buffers[best_index]," ,\n");
      numtokens=1;
      while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n")) 
         != NULL) numtokens++;

      if (debug>=10) {
        printf("  interpreting the line\n"); 
        printf("  best_index = %d\n",best_index); fflush(stdout);
        printf("  ptokens[0] = %s\n",ptokens[0]); 
        printf("  ptokens[1] = %s\n",ptokens[1]); 
        printf("  ptokens[2] = %s\n",ptokens[2]);
        if (numtokens == 4) {
          printf("  ptokens[3] = %s\n",ptokens[3]);
        }
        printf("  object_num = %d\n",object_num); fflush(stdout);
      }

      if (object_num>=MAX_NUM_OBJECTS) {
        printf("\n  Error:  attempt to define too many OBJECTS");
                printf(" (max = %d)\n", MAX_NUM_OBJECTS);
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }
      /* interpret the line */
      /* store from_node identifiers for this object */
      strcpy(object_map[object_num].from_node_label,ptokens[1]);
      strcpy(from_node_string,ptokens[1]);
      strcpy(object_map[object_num].to_node_label,ptokens[2]);

      if (debug>=10) {
        printf("  storing the label and parameters\n"); fflush(stdout);
      }

      /* store label for this object */
      if (!valid_label(ptokens[0])) {
        printf("  Error:  invalid label for BUFFER: %s\n", ptokens[0]);
        fflush(stdout);
        num_errors_found++;
      }
      strcpy(object_map[object_num].label,ptokens[0]);

      /* store parameters */
      if (buffer_num>=MAX_NUM_BUFFERS) {
        printf("\n  interpret_spk_file:  attempt to define too many buffers");
		printf(" (max = %d)\n", MAX_NUM_BUFFERS);
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }
      buffer[buffer_num].id = object_num;
      if (numtokens == 4) {
        buffer[buffer_num].t_delay = atof(ptokens[3]);
        if (!valid_number(ptokens[3])||buffer[buffer_num].t_delay<0) {
          printf("  Error:  invalid delay time for BUFFER: %s\n", 
            ptokens[3]);
          fflush(stdout);
          num_errors_found++;
        }
      } else {
        buffer[buffer_num].t_delay = 0.0;
      }

      if (debug>=10) {
        printf("  storing object type and sub_id\n"); fflush(stdout);
      }

      object_map[object_num].type = (int)BUFFER;
      object_map[object_num].sub_id = buffer_num;

      if (debug>=10) {
        printf("  searching inputs and neurons for matching node label\n");
        fflush(stdout);
      }

      /* search all inputs and neurons for matching node label */
      j=0;
      while (strcmp(from_node_string,object_map[j].to_node_label)!=0 &&
             j < input_num + neuron_num) j++;
      if (j < input_num + neuron_num) {
        from_node = j;
      } else {
        from_node = -1;
      }

      /* store parameters */
      buffer[buffer_num].id = object_num;
      buffer[buffer_num].object_from = from_node;

      if (debug>=10) {
        printf("%s\n",buffer_buffers[best_index]); fflush(stdout);
      }

      /* increment object_num, buffer_num */
      object_num++;
      buffer_num++;

      /* clear the line by copying last entry into used position */
      strcpy(buffer_buffers[best_index],buffer_buffers[i]);

    } /* END FOR all buffer lines */

    if (debug>=10) {
      printf("\n  sorting synapse strings:\n"); 
      printf("  Number of synapse strings: %d\n",num_synapses); fflush(stdout);
    }

    /* pull out the synapses */
    /* FOR all strings in the buffer */
    for (synapse_num=0;synapse_num<num_synapses;synapse_num++) {

      /* parse into tokens */
      ptokens[0] = (char *) strtok(buffer_synapses[synapse_num]," ,\n");
      numtokens=1;
      while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n")) 
              != NULL) numtokens++;

      /* interpret the line */

      /* parse out the to_node and from_node identifiers */
      strcpy(object_map[object_num].from_node_label,ptokens[0]);
      strcpy(from_node_string,ptokens[0]);
      strcpy(object_map[object_num].to_node_label,ptokens[1]);
      strcpy(to_node_string,ptokens[1]);

      /* find the corresponding object numbers */
      /* must search all inputs, neurons and buffers to find input for this */
	  /*   synapse */
      i=0;
      while (strcmp(from_node_string,object_map[i].to_node_label)!=0 &&
             i < input_num + neuron_num + buffer_num) i++;
      if (i < input_num + neuron_num + buffer_num) {
        from_node = i;
      } else {
        from_node = -1;   /* error condition, will catch it later */
      }

      /* only need to search neurons to find output for this synapse */
      i=input_num;
      while (strcmp(to_node_string,object_map[i].from_node_label)!=0 &&
             i < input_num + neuron_num) i++;
      if (i < input_num + neuron_num) {
        to_node = i;
      } else {
	    to_node = -1;
      }

      /* store parameters */
      if (synapse_num>=MAX_NUM_SYNAPSES) {
        printf("\n  interpret_spk_file:  attempt to define too many synapses ");
        printf("(max = %d)\n",MAX_NUM_SYNAPSES);
        printf("  Exiting.\n");
        fflush(stdout); exit(1);
      }
      synapse[synapse_num].id = object_num;
      synapse[synapse_num].object_from = from_node;
      synapse[synapse_num].object_to = to_node;
      synapse[synapse_num].i_weight = atof(ptokens[2]);
      if (!valid_number(ptokens[2])) {
        printf("  Error:  invalid weight current for SYNAPSE: %s\n", 
          ptokens[2]);
        fflush(stdout);
        num_errors_found++;
      }
      synapse[synapse_num].t_neuro = atof(ptokens[3]);
      if (!valid_number(ptokens[3])||synapse[synapse_num].t_neuro<0) {
        printf("  Error:  invalid post-synaptic current pulse time for SYNAPSE: %s\n", 
          ptokens[3]);
        fflush(stdout);
        num_errors_found++;
      }
      synapse[synapse_num].cliplevel = 1;  /* default value */
      if (numtokens==5) {
        synapse[synapse_num].cliplevel = atoi(ptokens[4]);
        if (!valid_number(ptokens[4])||synapse[synapse_num].cliplevel<1) {
          printf("  Error:  invalid clip level for SYNAPSE: %s\n", 
            ptokens[4]);
          fflush(stdout);
          num_errors_found++;
        }
      } 

      object_map[object_num].type = (int)SYNAPSE;
      object_map[object_num].sub_id = synapse_num;

      /* generate the object label for the synapse */
      strcpy(object_map[object_num].label,"s");
      strcat(object_map[object_num].label,object_map[from_node].label);
      strcat(object_map[object_num].label,object_map[to_node].label);

      /* increment object_num */
      object_num++;

    } /* END FOR all synapse lines */

    /* check for duplicate synapse labels */
    /* FOR all strings in the buffer */
    for (i=num_synapses-1;i>=0;i--) {
      duplicate_found = FALSE;
      suffix_string[0] = 'a';
      suffix_string[1] = '\0';
      for (j=0;j<i;j++) {
        if (strcmp(object_map[synapse[i].id].label,
                   object_map[synapse[j].id].label)==0) {
          duplicate_found = TRUE;
          strcat(object_map[synapse[j].id].label,suffix_string);
          suffix_string[0] += 1;
        }
      }
      if (duplicate_found) {
        strcat(object_map[synapse[i].id].label,suffix_string);
      }
    }

    /* close files */
    fclose(spk_file);
    fclose(spk2_file);
    if (debug <10)  system("rm tempspk2\n");

/* END */
}


/*****************************************************************/
void
interpret_sim_file()
/*****************************************************************/
{

    extern struct object_map_struct object_map[];
    extern struct neuron_struct neuron[];
    extern struct synapse_struct synapse[];
    extern int          neuron_num,synapse_num,input_num,object_num;
    extern int          debug, displayrefract, output_format;
    extern int          max_mathlabel_length;
    extern float        start_runtime;
    extern float        threshold;
    extern float        runtime;
    extern char        *object_type_string[];
    extern FILE        *sim_file;
    extern FILE        *evt_file;
    extern char         sim_file_name[];
    extern char         evt_file_name[];


    FILE               *fopen();
    FILE               *temp1_file;
    char                network_spec[MAX_CHARS_PER_LINE];
    char                network_spec_copy[MAX_CHARS_PER_LINE];
    char                dummy_line[MAX_CHARS_PER_LINE];
    char                command_string[MAX_CHARS_PER_LINE];
    char               *ptokens[MAX_TOKENS_PER_LINE];
    int                 ii,i,j,numtokens;
    int                 line_counter;
    int                 display_counter;
    int                 valid_command_found;
    int                 pulse_object;
    int                 display_type;
    long                seed, seedgauss;
    float               start_time, finish_time, period, lambda, time;
    float               interspike_time, burst_period, burst_start_time;
    int                 num_spikes_per_burst;
    int                 gaussjitter;
    float               jitter_time, stdev;

    /* open files */
    if ((sim_file = fopen(sim_file_name,"r"))==NULL) {
      printf("\n  ERROR:  unable to find file %s\n",sim_file_name);
      printf("  Exiting.\n");
      fflush(stdout); exit(1);
    }

    if ((evt_file = fopen(evt_file_name,"w+"))==NULL) {
      printf("\n  interpret_sim_file:  unable to open file:  %s\n",evt_file_name);
      printf("  Exiting.\n");
      fflush(stdout); exit(1);
    }

    if ((temp1_file = fopen("temp1","w+"))==NULL) {
      printf("\n  interpret_sim_file:  unable to open file:  temp1\n");
      printf("  Exiting.\n");
      fflush(stdout); exit(1);
    }

    /* initialize basic values */
    line_counter = 0;
    display_counter = 0;

    /* set display_order to simple default pattern (all traces displayed) */
    display_type = (int)DEFAULT;
    for (i=0;i<object_num;i++) {
      object_map[i].display_order = i;
    }

    if (debug>=5) {
      printf("  lines in .sim file:\n"); fflush(stdout);
    }

    /* WHILE more lines exist */
    while (fgets(network_spec,MAX_CHARS_PER_LINE,sim_file) != NULL) {
 
      if (debug>=5) {
        printf("    %s",network_spec); fflush(stdout);
      }

      /* increment line_counter */
      line_counter++;
      valid_command_found = FALSE;

      /* read a line */
      strcpy(network_spec_copy,network_spec);

      /* parse into tokens */
      ptokens[0] = (char *) strtok(network_spec," ,\n");

      if (ptokens[0] != NULL) {

        numtokens=1;
        while ((ptokens[numtokens] = (char *) strtok((char *)NULL," ,\n")) 
                != NULL) numtokens++;

        /* (OBSOLETE, superceded by "display refract") */
        /* IF first token is "displayrefract" */
        if (strcmp(ptokens[0],"displayrefract")==0) {

          /* set displayrefract = 1 */
          displayrefract = 1;
          valid_command_found = TRUE;
          printf("\n  Warning:  OBSOLETE displayrefract command found.\n");
          printf("            please use \"display refract\" in your .sim file\n\n");
          fflush(stdout); 

        /* END IF first token is "displayrefract" */
        }


        /* (OBSOLETE, superceded by "display mathematica") */
        /* IF first token is "mathematica" */
        if (strcmp(ptokens[0],"mathematica")==0) {

          /* set output_format = MATHEMATICA */
          output_format = (int)MATHEMATICA;
          valid_command_found = TRUE;

        /* END IF first token is "mathematica" */
        }

        /* IF first token is "display" */
        if (strcmp(ptokens[0],"display")==0) {

          /* IF second token is "all" */
          if (strcmp(ptokens[1],"all")==0) {

            /* IF display_type=default or all */
            if (display_type==(int)DEFAULT || display_type == (int)ALL) {

              /* set display_type = all */
              display_type=(int)ALL;
              valid_command_found = TRUE;

            /* ELSE die */
            } else {

              printf("\n  interpret_sim_file:  duplicate display commands.\n");
              printf("  Exiting.\n");
              fflush(stdout); exit(1);
            }

          /* END IF second token is "all" */
          }

          /* IF second token is "refract" */
          if (strcmp(ptokens[1],"refract")==0) {

            /* set displayrefract = 1 */
            displayrefract = 1;
            valid_command_found = TRUE;

          /* END IF second token is "refract" */
          }

          /* IF second token is "mathematica" */
          if (strcmp(ptokens[1],"mathematica")==0) {

            /* set output_format = MATHEMATICA */
            output_format = (int)MATHEMATICA;
            valid_command_found = TRUE;

          /* END IF second token is "mathematica" */
          }

          /* IF second token is "xvgr" */
          if (strcmp(ptokens[1],"xvgr")==0) {

            /* set output_format = XVGR */
            output_format = (int)XVGR;
            valid_command_found = TRUE;

          /* END IF second token is "xvgr" */
          }

          /* IF second token is "gnuplot" */
          if (strcmp(ptokens[1],"gnuplot")==0) {

            /* set output_format = GNUPLOT */
            output_format = (int)GNUPLOT;
            valid_command_found = TRUE;

          /* END IF second token is "gnuplot" */
          }

          /* IF second token is "cview" */
          if (strcmp(ptokens[1],"cview")==0) {

            /* set output_format = CVIEW */
            output_format = (int)CVIEW;
            valid_command_found = TRUE;

          /* END IF second token is "cview" */
          }

          /* IF second token is "max_label_length" */
          if (strcmp(ptokens[1],"max_label_length")==0) {

            max_mathlabel_length = atoi(ptokens[2]);
            valid_command_found = TRUE;

          /* END IF second token is "refract" */
          }

          /* IF second token is "nosynapse" */
          if (strcmp(ptokens[1],"nosynapse")==0) {

            /* IF display_type=default or nosynapse, */
            if (display_type==(int)DEFAULT || display_type == (int)NOSYNAPSE) {

              /* set display_type = nosynapse */
              display_type=(int)NOSYNAPSE;
              valid_command_found = TRUE;

              /* set all synapse display_orders to NO_DISPLAY_VALUE */
              for (i=0;i<synapse_num;i++) {
                object_map[synapse[i].id].display_order = NO_DISPLAY_VALUE;
              }        
  
            /* ELSE die */
            } else {
              printf("\n  interpret_sim_file:  duplicate display commands.\n");
              printf("  Exiting.\n");
              fflush(stdout); exit(1);
            }

          /* END IF */
          }

          /* IF second token is "list" */
          if (strcmp(ptokens[1],"list")==0) {

            /* IF display_type=default */
            if (display_type==(int)DEFAULT) {

              /* set display_type = list */
              display_type=(int)LIST;
              valid_command_found = TRUE;

              /* clear all display_orders to NO_DISPLAY_VALUE */
              for (i=0;i<object_num;i++) {
                object_map[i].display_order = NO_DISPLAY_VALUE;
              }

              /* set display_orders according to list */
              for (i=2;i<numtokens;i++) {
                j=0;
                while (strcmp(object_map[j].label,ptokens[i])!=0 && 
                        j<object_num) j++;

                if (j<object_num) {
                  object_map[j].display_order = display_counter;
                  object_map[j].saved_file_num = display_counter;
                  display_counter++;
              
                } else {

                  /* die */
                  printf("\n  interpret_sim_file:  %s: no such object to display.\n",ptokens[i]);
                  printf("  Exiting.\n");
                  fflush(stdout); exit(1);
                }
              }

            /* ELSE */
            } else {

              /* IF display_type=list */
              if (display_type==(int)LIST) {

                valid_command_found = TRUE;

                /* set display_orders according to list */
                for (i=2;i<numtokens;i++) {
                  j=0;
                  while (strcmp(object_map[j].label,ptokens[i])!=0 && 
                          j<object_num) j++;

                  if (j<object_num) {
                    object_map[j].display_order = display_counter;
                    object_map[j].saved_file_num = display_counter;
                    display_counter++;
              
                  } else {

                    /* die */
                    printf("\n  interpret_sim_file:  %s: no such object to display.\n",ptokens[i]);
                    printf("  Exiting.\n");
                    fflush(stdout); exit(1);
                  }
                }

              /* ELSE */
              } else {

                /* die */
                printf("\n  interpret_sim_file:  duplicate display commands.\n");
                printf("  Exiting.\n");
                fflush(stdout); exit(1);

              /* END IF */
              }
            /* END IF */
            }
          /* END IF */
          }

          if (!valid_command_found) {
            printf("\n  interpret_sim_file:  invalid display command: %s\n",
              network_spec_copy);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }

        /* END IF first token is "display" */
        }

        /* IF first token is "end_runtime" */
        if (strcmp(ptokens[0],"end_runtime")==0) {

          /* set runtime level */
          runtime = atof(ptokens[1]);
          if (runtime<0) {
            printf("\n  interpret_sim_file:  runtime < 0\n");
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          valid_command_found = TRUE;
        }
        /* END IF first token is "end_runtime" */
        /* IF first token is "start_runtime" */
        if (strcmp(ptokens[0],"start_runtime")==0) {
  
          /* set runtime level */
          start_runtime = atof(ptokens[1]);
          if (start_runtime<0) {
            printf("\n  interpret_sim_file:  start_runtime < 0\n");
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          valid_command_found = TRUE;
        }
        /* END IF first token is "start_runtime" */
  
        /* IF first token is "threshold" */
        if (strcmp(ptokens[0],"threshold")==0) {
  
          /* set runtime level */
          threshold = atof(ptokens[1]);
          if (threshold<0) {
            printf("\n  interpret_sim_file:  threshold < 0\n");
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          valid_command_found = TRUE;
        }
        /* END IF first token is "threshold" */
  
        /* IF first token is "spike" */
        if (strcmp(ptokens[0],"spike")==0) {
  
          /* find object with the given label */
          i=0;
          while (strcmp(ptokens[1],object_map[i].label)!=0 && i<input_num) i++;
          if (i<input_num) {
            pulse_object = i;
          } else {
            printf("  interpret_sim_file:  spike command given with bad object label:  %s\n",ptokens[1]);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
  
          /* copy line to temp1 file with object number */
          fprintf(temp1_file,"%d %s\n",pulse_object,ptokens[2]);
          fflush(temp1_file);
  
          valid_command_found = TRUE;
        }
        /* END IF first token is "spike" */
  
  
        /* IF first token is "train" */
        if (strcmp(ptokens[0],"train")==0) {
  
          /* find object with the given label */
          i=0;
          while (strcmp(ptokens[1],object_map[i].label)!=0 && i<input_num) i++;
          if (i<input_num) {
            pulse_object = i;
          } else {
            printf("  interpret_sim_file:  train command given with bad object label:  %s\n",ptokens[1]);
            printf("  Exiting.\n");
            fflush(stdout);
            exit(1);
          }

          start_time = atof(ptokens[2]);
          finish_time = atof(ptokens[3]);
          period = atof(ptokens[4]);
          gaussjitter = FALSE; /* default */
          if (numtokens==7) {  /* optional parameters */
            gaussjitter = TRUE; 
            stdev = atof(ptokens[5]);
            seedgauss = atol(ptokens[6]);
          }
  
          if (period <= 0) {
            printf("  interpret_sim_file:  period <= 0 for spike train, object label:  %s\n",ptokens[1]);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          if (start_time < 0) {
            printf("  interpret_sim_file:  start_time < 0 for spike train, object label:  %s\n",ptokens[1]);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          if (finish_time < start_time) {
            printf("  interpret_sim_file:  finish_time < start_time for spike train,\n                     object label:  %s\n",ptokens[1]);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
  
          if (gaussjitter) {
            time = start_time;
            while (time <= finish_time) {
              /* copy line to temp1 file with object number */
              fprintf(temp1_file,"%d %f\n",pulse_object,time);
              fflush(temp1_file);
              /* compute jitter_time, don't accept values less than -period */
              while ((jitter_time = gasdev(&seedgauss)*stdev) < -period) ;
              time += period + jitter_time;
            }
          } else {
            i=0;
            while ((time = start_time + i*period) <= finish_time) {
              /* copy line to temp1 file with object number */
              fprintf(temp1_file,"%d %f\n",pulse_object,time);
              fflush(temp1_file);
              i++;
            }
          }
          valid_command_found = TRUE;
        }
        /* END IF first token is "train" */
  
        /* IF first token is "poisson" */
        if (strcmp(ptokens[0],"poisson")==0) {
  
          /* find object with the given label */
          i=0;
          while (strcmp(ptokens[1],object_map[i].label)!=0 && i<input_num) i++;
          if (i<input_num) {
            pulse_object = i;
          } else {
            printf("  interpret_sim_file:  poisson command given with bad object label:  %s\n",ptokens[1]);
            printf("  Exiting.\n");
            fflush(stdout);
            exit(1);
          }

          start_time = atof(ptokens[2]);
          finish_time = atof(ptokens[3]);
          lambda = atof(ptokens[4]);
          seed = atol(ptokens[5]);
  
          if (lambda <= 0) {
            printf("  interpret_sim_file:  lambda <= 0 for poisson spike train, object label:  %s\n",ptokens[1]);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          if (start_time < 0) {
            printf("  interpret_sim_file:  start_time < 0 for spike train, object label:  %s\n",ptokens[1]);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          if (finish_time < start_time) {
            printf("  interpret_sim_file:  finish_time < start_time for spike train,\n                     object label:  %s\n",ptokens[1]);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
  
          time = start_time;
          while (time <= finish_time) {
            /* copy line to temp1 file with object number */
            fprintf(temp1_file,"%d %f\n",pulse_object,time);
            fflush(temp1_file);
            if (debug>=5) {
              printf("  poisson random spike at time %f \n",time);
            }
            time += -lambda*log(ran0(&seed));
          }
          valid_command_found = TRUE;
        }
        /* END IF first token is "poisson" */

  
        /* IF first token is "burst" */
        if (strcmp(ptokens[0],"burst")==0) {
  
          /* find object with the given label */
          i=0;
          while (strcmp(ptokens[1],object_map[i].label)!=0 && i<input_num) i++;
          if (i<input_num) {
            pulse_object = i;
          } else {
            printf("  interpret_sim_file:  burst command given with bad object label:  %s\n",ptokens[1]);
            printf("  Exiting.\n");
            fflush(stdout);
            exit(1);
          }

          start_time = atof(ptokens[2]);
          finish_time = atof(ptokens[3]);
          interspike_time = atof(ptokens[4]);
          num_spikes_per_burst = atoi(ptokens[5]);
          burst_period = atof(ptokens[6]);
  
          if (start_time < 0) {
            printf("  interpret_sim_file:  start_time < 0 for spike burst, object label:  %s\n",ptokens[1]);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          if (finish_time < start_time) {
            printf("  interpret_sim_file:  finish_time < start_time for spike burst,\n                     object label:  %s\n",ptokens[1]);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          if (interspike_time <= 0) {
            printf("  interpret_sim_file:  inter-spike time <= 0 for spike burst, object label:  %s\n",ptokens[1]);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          if (num_spikes_per_burst <= 0) {
            printf("  interpret_sim_file:  number of spikes per burst <= 0 for spike burst, object label:  %s\n",ptokens[1]);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
          if (burst_period <= interspike_time*num_spikes_per_burst) {
            printf("  interpret_sim_file:  burst period less than the interspike time \n    times the number of spikes per burst,\n   object label:  %s\n",ptokens[1]);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
  
          burst_start_time = start_time;
          ii=0;
          while ((burst_start_time=start_time+ii*burst_period)<=finish_time) {
            i=0;
            while ((time = burst_start_time + i*interspike_time) <= finish_time &&
                    i<num_spikes_per_burst) {
              /* copy line to temp1 file with object number */
              fprintf(temp1_file,"%d %f\n",pulse_object,time);
              fflush(temp1_file);
              if (debug>=5) {
                printf("  burst spike at time %f \n",time);
              }
              i++;
            }
            ii++;
          }
          valid_command_found = TRUE;
        }
        /* END IF first token is "burst" */
  

        /* check for comment */
        if (!valid_command_found) {
          strcpy(dummy_line,ptokens[0]);
          if (dummy_line[0]!='#') {
            printf("\n  interpret_sim_file:  unrecognized command in line %d of file %s\n",
               line_counter,sim_file_name);
            printf("  unrecognized command:  %s\n",network_spec_copy);
            printf("  Exiting.\n");
            fflush(stdout); exit(1);
          }
        }
      }
    } /* END WHILE more lines exist */

    /* close files */
    fclose(temp1_file);
    fclose(sim_file);

    /* sort the temp1 file on the basis of event time to <file>.evt */
    /* sprintf(command_string,"sort +2.0b -n -o %s temp1\n",evt_file_name); */
    sprintf(command_string,"sort +1n -2 -o %s temp1\n",evt_file_name);
    system(command_string);

    if (debug<10) {
      remove("temp1");
    }

/* END */
}
/*****************************************************************/
float
ran0(idum)
/*    random number generator from Numerical Recipes, 2nd ed.    */
/*****************************************************************/
long *idum;
{
    long k;
    float ans;

    *idum ^= MASK;
    k=(*idum)/IQ;
    *idum=IA*(*idum-k*IQ)-IR*k;
    if (*idum < 0) *idum += IM;
    ans=AM*(*idum);
    *idum ^= MASK;
    return ans;
}

/*****************************************************************/
float
gasdev(idum)
/*    gaussian distributed random number, from Numerical Recipes */
/*****************************************************************/
long *idum;
{
    extern int iset;
    extern float gset;
    float fac, r, v1, v2;
    float ran0();

    if (iset==0) {
      do {
        v1=2.0*ran0(idum)-1.0;
        v2=2.0*ran0(idum)-1.0;
        r=v1*v1+v2*v2;
      } while (r>=1.0||r==0.0);
      fac=sqrt(-2.0*log(r)/r);
      gset=v1*fac;
      iset=1;
      return v2*fac;
    } else {
      iset=0;
      return gset;
    }
}

/*****************************************************************/
int
valid_label(p)
/*    checks that a string is a valid object label */
/*****************************************************************/
char *p;
{
    int i, result;

    /* presumed guilty */
    result = FALSE;

    /* IF first char is an alpha */
    if (isalpha((int)*p)) {

      result = TRUE;
      p++;
      i=1;
      while (*p!='\0' && i<MAX_CHARS_PER_LINE) {
        if (!isalnum((int)*p)) result=FALSE;
        p++;
        i++;
      }
    }
    return result;
}
/*****************************************************************/
int
valid_number(p)
/*    checks that a string is a valid number */
    /* first char must be digit,+,-, or . */
    /* remaining chars must be either digit or . */
    /* only one . allowed */
    /* must be at least one digit present */
/*****************************************************************/
char *p;
{
    int i, result, num_decimal_points, num_digits;

    i=0;
    num_decimal_points=0;
    num_digits = 0;
    result = TRUE;

    while (*p!='\0' && i<MAX_CHARS_PER_LINE && result == TRUE) {
      switch(*p) {
        case '.':
          if (num_decimal_points==0) {
            num_decimal_points++;
          } else {
            result = FALSE;
          }
          break;
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
          num_digits++;
          break;
        case '+':
        case '-':
          if (i!=0) {
            result = FALSE;
          } 
          break;
        default:
          result = FALSE;
          break;
      }
      i++;
      p++;
    }
    if (num_digits == 0) result = FALSE;
    return result;
}
/*****************************************************************/
int
found_node_string(p)
/*    looks for a string whose first four characters are "node" */
/*****************************************************************/
char *p;
{
    int i, result;

    result = TRUE;
    i=0;
    while (*p!='\0' && i<MAX_CHARS_PER_LINE && result==TRUE) {
      if (*p!='n'&&i==0) result=FALSE;
      if (*p!='o'&&i==1) result=FALSE;
      if (*p!='d'&&i==2) result=FALSE;
      if (*p!='e'&&i==3) result=FALSE;
      p++;
      i++;
    }
    if (i<4) result=FALSE;
    return result;
}
